"""Factory for getting cost functions and their initializers from strings.

Recipe for adding new costs:
    1. Add a new module in skchange.costs: <cost_name>_cost.py
    2. Add two functions:
        init_<cost_name>_cost, which precomputes quantities that should be precomputed,
        often partial sums and such.
        <cost_name>_cost, which takes in the output of init_<cost_name>_cost as its
        first argument, and starts and ends indices of the costs to be computed as its
        second and third arguments.
    3. Add the name of the cost to VALID_COSTS below.
    4. Add the name of the cost to the docstring of cost_factory below.
    4. Add a new if-statement in cost_factory below.

"""

from skchange.costs.mean_cost import init_mean_cost, mean_cost

VALID_COSTS = ["mean"]


def cost_factory(cost_name: str):
    """Return cost function and its initializer.

    Parameters
    ----------
    cost_name : str
        Name of cost function. Must be one of 'mean'.

    Returns
    -------
    cost_func : Callable
        Cost function.
    init_cost_func : Callable
        Cost function initializer.
    """
    if cost_name == "mean":
        return mean_cost, init_mean_cost
    else:
        message = (
            f"cost_name={cost_name} not recognized."
            + f" Must be one of {', '.join(VALID_COSTS)}"
        )
        raise ValueError(message)



"""Numba-optimized functions for calculating various statistics."""

import numpy as np
from numba import njit


@njit(cache=True)
def col_cumsum(x: np.ndarray):
    """Calculate the cumulative sum of each column in a 2D array.

    Parameters
    ----------
    x : np.ndarray
        2D array.
    """
    cumsum = np.zeros_like(x)
    for j in range(x.shape[1]):
        cumsum[:, j] = np.cumsum(x[:, j])
    return cumsum



"""Test statistic for differences in the mean."""

from typing import Tuple

import numpy as np
from numba import njit

from skchange.utils.numba.stats import col_cumsum


@njit(cache=True)
def init_mean_score(X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Precompute sums for mean_score.

    Parameters
    ----------
    X : np.ndarray
        2D array.

    Returns
    -------
    sums : np.ndarray
        Cumulative sums of X.
    """
    n = X.shape[0]
    p = X.shape[1]
    # 0.0 as first row to make calculations work also for start = 0 in mean_score
    sums = np.zeros((n + 1, p))
    sums[1:] = col_cumsum(X)
    return sums


@njit(cache=True)
def mean_score(
    precomputed_params: np.ndarray,
    start: np.ndarray,
    end: np.ndarray,
    split: np.ndarray,
) -> np.ndarray:
    """Calculate the CUSUM score for a change in the mean.

    Compares the mean of the data before and after the split within the interval from
    start:end (both inclusive).

    Parameters
    ----------
    precomputed_params : np.ndarray
        Precomputed parameters from init_mean_score.
    start : np.ndarray
        Start indices of the intervals to test for a change in the mean.
    end : np.ndarray
        End indices of the intervals to test for a change in the mean.
    split : np.ndarray
        Split indices of the intervals to test for a change in the mean.

    Returns
    -------
    score : np.ndarray
        Scores for a difference in the mean at the given intervals and splits

    Notes
    -----
    To optimize performance, no checks are performed on (start, split, end).
    """
    sums = precomputed_params
    before_sum = sums[split + 1] - sums[start]
    before_weight = np.sqrt((end - split) / ((end - start + 1) * (split - start + 1)))
    before_weight = before_weight.reshape(-1, 1)
    after_sum = sums[end + 1] - sums[split + 1]
    after_weight = np.sqrt((split - start + 1) / ((end - start + 1) * (end - split)))
    after_weight = after_weight.reshape(-1, 1)
    return np.sum(np.abs(after_weight * after_sum - before_weight * before_sum), axis=1)


@njit(cache=True)
def mean_anomaly_score(
    precomputed_params: np.ndarray,
    interval_start: np.ndarray,
    interval_end: np.ndarray,
    anomaly_start: np.ndarray,
    anomaly_end: np.ndarray,
) -> np.ndarray:
    """Calculate the CUSUM score for difference in the mean of a subinterval.

    Compares the mean of the data in anomaly_start:anomaly_end (both inclusive)
    to the mean of the complement of interval_start:interval_end (both inclusive) and
    anomaly_start:anomaly_end.

    The overall and anomalous intervals must satisfy
    `interval_start > anomaly_start <= anomaly_end <= interval_end`

    Parameters
    ----------
    precomputed_params : np.ndarray
        Precomputed parameters from init_mean_score.
    interval_start : np.ndarray
        Start indices of the intervals to test for an anomaly in.
    interval_end : np.ndarray
        End indices of the intervals to test for an anomaly in.
    anomaly_start : np.ndarray
        Start indices of the anomalies.
    anomaly_end : np.ndarray
        End indices of the anomalies.

    Returns
    -------
    score : float
        Score for a difference in the mean.

    Notes
    -----
    To optimize performance, no checks are performed on the inputs.
    """
    sums = precomputed_params
    baseline_sum = (
        sums[interval_end + 1]
        - sums[anomaly_end + 1]
        + sums[anomaly_start]
        - sums[interval_start]
    )
    baseline_n = interval_end - anomaly_end + anomaly_start - interval_start
    baseline_n = baseline_n.reshape(-1, 1)
    baseline_mean = baseline_sum / baseline_n
    anomaly_sum = sums[anomaly_end + 1] - sums[anomaly_start]
    anomaly_n = (anomaly_end - anomaly_start + 1).reshape(-1, 1)
    anomaly_mean = anomaly_sum / anomaly_n
    weight = (1 / anomaly_n + 1 / baseline_n) ** (-1 / 2)
    return np.sum(np.abs(weight * (baseline_mean - anomaly_mean)), axis=1)



import numpy as np
import plotly.express as px

from skchange.anomaly_detectors.moscore_anomaly import MoscoreAnomaly
from skchange.datasets.generate import generate_anomalous_data
from skchange.scores.mean_score import init_mean_score, mean_anomaly_score
from skchange.utils.benchmarking.profiler import Profiler

n = 500
df = generate_anomalous_data(
    n, anomalies=[(100, 119), (250, 299)], means=[10.0, 5.0], random_state=1
)
px.scatter(df)

detector = MoscoreAnomaly(
    score="meanvar",
    min_anomaly_length=10,
    max_anomaly_length=100,
    threshold_scale=3.0,
    left_bandwidth=50,
)
anomalies = detector.fit_predict(df)

detector = MoscoreAnomaly(
    score="mean",
    min_anomaly_length=10,
    max_anomaly_length=1000,
    left_bandwidth=20,
    labels="score",
)
scores = detector.fit_predict(df)
scores["length"] = scores["anomaly_end"] - scores["anomaly_start"] + 1
px.scatter(scores, x="anomaly_start", y="score", color="length")


# Profiling
n = int(1e5)
df = generate_anomalous_data(n, anomalies=[(100, 119), (250, 299)], means=[10.0, 5.0])
detector = MoscoreAnomaly(
    score="mean",
    min_anomaly_length=10,
    max_anomaly_length=100,
    left_bandwidth=50,
)

profiler = Profiler()
profiler.start()
anomalies = detector.fit_predict(df)
profiler.stop()


# Check mean anomaly score
n = 500
df = generate_anomalous_data(n, anomalies=[(100, 119), (250, 299)], means=[10.0, 5.0])
df = df.iloc[:40]

params = init_mean_score(df.values)
starts = np.arange(10, 20)
ends = starts + 10 - 1
background_starts = starts - 10
background_ends = ends + 10
scores = mean_anomaly_score(params, background_starts, background_ends, starts, ends)



"""Data generators."""

from numbers import Number
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from scipy.stats import multivariate_normal
from sktime.annotation.datagen import piecewise_normal_multivariate


def generate_teeth_data(
    n_segments: int,
    segment_length: int,
    p: int = 1,
    mean: float = 0.0,
    variance: float = 1.0,
    covariances: np.ndarray = None,
    affected_proportion: float = 1.0,
    random_state: int = None,
) -> pd.DataFrame:
    """
    Generate a DataFrame with a teeth-shaped or square wave mean signal.

    Parameters
    ----------
        n_segments : int
            Number of segments to generate.
        segment_length : int
            Length of each segment.
        p : int, optional (default=1)
            Number of dimensions.
        mean : float, optional (default=0.0)
            Mean of each alternating segment.
        variance : float, optional (default=1.0)
            Variances of each alternating segment.
        covariances : array-like, optional (default=None)
            Covariances between dimensions.
        affected_proportion : float, optional (default=1.0)
            Proportion of components {1, ..., p} that are affected by each change.
        random_state : int or RandomState, optional
            Seed or random state for reproducible results. Defaults to None.

    Returns
    -------
        pd.DataFrame: DataFrame with generate_teeth_data-shaped segments.
    """
    means = []
    vars = []
    n_affected = int(np.round(p * affected_proportion))
    for i in range(n_segments):
        zero_mean = [0] * p
        changed_mean = [mean] * n_affected + [0] * (p - n_affected)
        mean_vec = zero_mean if i % 2 == 0 else changed_mean
        means.append(mean_vec)
        one_var = [1] * p
        changed_var = [variance] * n_affected + [1] * (p - n_affected)
        vars_vec = one_var if i % 2 == 0 else changed_var
        vars.append(vars_vec)

    segment_lengths = [segment_length] * n_segments
    x = piecewise_normal_multivariate(
        means, segment_lengths, vars, covariances, random_state
    )
    df = pd.DataFrame(x, index=range(len(x)))
    return df


def generate_anomalous_data(
    n: int = 100,
    anomalies: Union[Tuple[int, int], List[Tuple[int, int]]] = (71, 80),
    means: Union[float, List[float], List[np.ndarray]] = 3.0,
    variances: Union[float, List[float], List[np.ndarray]] = 1.0,
    random_state: int = None,
) -> pd.DataFrame:
    """
    Generate multivariate normal data with anomalies.

    Parameters
    ----------
        n : int, optional (default=100)
            Number of observations.
        anomalies : list of tuples, optional (default=[(71, 80)])
            List of tuples of the form (start, end) indicating the start and end of an
            anomaly.
        means : list of floats or list of arrays, optional (default=[0.0])
            List of means for each segment.
        variances : list of floats or list of arrays, optional (default=[1.0])
            List of variances for each segment.
        random_state : int or RandomState, optional
            Seed or random state for reproducible results. Defaults to None.

    Returns
    -------
        pd.DataFrame: DataFrame with generate_teeth_data-shaped segments.
    """
    if isinstance(anomalies, Tuple):
        anomalies = [anomalies]
    if isinstance(means, Number):
        means = [means]
    if isinstance(variances, Number):
        variances = [variances]

    if isinstance(means[0], Number):
        means = [np.array([mean]) for mean in means]
    if isinstance(variances[0], Number):
        variances = [np.array([variance]) for variance in variances]

    if len(means) == 1:
        means = means * len(anomalies)
    if len(variances) == 1:
        variances = variances * len(anomalies)

    if len(anomalies) != len(means) or len(anomalies) != len(variances):
        raise ValueError("Number of anomalies, means and variances must be the same.")
    if any([len(anomaly) != 2 for anomaly in anomalies]):
        raise ValueError("Anomalies must be of length 2.")
    if any([anomaly[0] > anomaly[1] for anomaly in anomalies]):
        raise ValueError("The start of an anomaly must be before its end.")
    if any([anomaly[1] > n - 1 for anomaly in anomalies]):
        raise ValueError("Anomalies must be within the range of the data.")

    p = len(means[0])
    x = multivariate_normal.rvs(np.zeros(p), np.eye(p), n, random_state)
    for anomaly, mean, variance in zip(anomalies, means, variances):
        start, end = anomaly
        x[start : end + 1] = mean + np.sqrt(variance) * x[start : end + 1]
    df = pd.DataFrame(x, index=range(len(x)))
    return df


def add_linspace_outliers(df, n_outliers, outlier_size):
    """
    Add outliers to a DataFrame at evenly spaced positions.

    Parameters
    ----------
        df : pd.DataFrame
            DataFrame to add outliers to.
        n_outliers : int
            Number of outliers to add.
        outlier_size : float
            Size of the outliers.

    Returns
    -------
        pd.DataFrame: DataFrame with outliers added.
    """
    outlier_positions = np.linspace(0, df.size - 1, n_outliers, dtype=int)
    df.iloc[outlier_positions] += outlier_size
    return df



"""Tests for MoscoreAnomaly and all available anomaly scores."""

import numpy as np
import pytest

from skchange.anomaly_detectors.moscore_anomaly import MoscoreAnomaly
from skchange.datasets.generate import generate_anomalous_data
from skchange.scores.score_factory import VALID_ANOMALY_SCORES

true_anomalies = [(50, 59), (120, 129)]
anomaly_data = generate_anomalous_data(
    200, anomalies=true_anomalies, means=[10.0, 5.0], random_state=5
)


@pytest.mark.parametrize("score", VALID_ANOMALY_SCORES)
def test_moscore_anomalies(score):
    """Test Moscore anomalies."""
    detector = MoscoreAnomaly.create_test_instance()
    detector.set_params(score=score, fmt="sparse", labels="int_label")
    anomalies = detector.fit_predict(anomaly_data)
    assert len(anomalies) == len(true_anomalies)
    for i, (start, end) in enumerate(true_anomalies):
        assert anomalies.loc[i, "start"] == start and anomalies.loc[i, "end"] == end


@pytest.mark.parametrize("score", VALID_ANOMALY_SCORES)
def test_moscore_scores(score):
    """Test Moscore scores."""
    detector = MoscoreAnomaly.create_test_instance()
    detector.set_params(score=score, fmt="sparse", labels="int_label")
    scores = detector.fit_predict(anomaly_data)
    assert np.all(scores >= 0.0)


@pytest.mark.parametrize("score", VALID_ANOMALY_SCORES)
def test_moscore_tuning(score):
    """Test Moscore tuning."""
    detector = MoscoreAnomaly.create_test_instance()
    detector.set_params(
        score=score, threshold_scale=None, fmt="dense", labels="indicator"
    )
    detector.fit(anomaly_data)
    assert detector.threshold_ > 0.0



"""The collective and point anomalies (CAPA) algorithm."""
__author__ = ["mtveten"]
__all__ = ["Capa"]

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.anomaly_detectors.mvcapa import dense_capa_penalty, run_base_capa
from skchange.anomaly_detectors.utils import format_anomaly_output
from skchange.costs.saving_factory import saving_factory
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_larger_than


@njit
def run_capa(
    X: np.ndarray,
    saving_func: Callable,
    saving_init_func: Callable,
    collective_alpha: float,
    point_alpha: float,
    min_segment_length: int,
    max_segment_length: int,
) -> Tuple[np.ndarray, List[Tuple[int, int]], List[Tuple[int, int]]]:
    params = saving_init_func(X)
    collective_betas = np.zeros(1)
    point_betas = np.zeros(1)
    return run_base_capa(
        X,
        params,
        saving_func,
        collective_alpha,
        collective_betas,
        point_alpha,
        point_betas,
        min_segment_length,
        max_segment_length,
    )


class Capa(BaseSeriesAnnotator):
    """Collective and point anomaly detection.

    An efficient implementation of the CAPA algorithm [1]_ for anomaly detection.
    It is implemented using the 'savings' formulation of the problem given in [2]_.

    Capa can be applied to both univariate and multivariate data, but does not infer
    the subset of affected components for each anomaly in the multivariate case. See the
    Mvcapa class if such inference is desired.

    Parameters
    ----------
    saving : str (default="mean")
        Saving function to use for anomaly detection.
    collective_penalty_scale : float, optional (default=2.0)
        Scaling factor for the collective penalty.
    point_penalty_scale : float, optional (default=2.0)
        Scaling factor for the point penalty.
    min_segment_length : int, optional (default=2)
        Minimum length of a segment.
    max_segment_length : int, optional (default=1000)
        Maximum length of a segment.
    ignore_point_anomalies : bool, optional (default=False)
        If True, detected point anomalies are not returned by .predict(). I.e., only
        collective anomalies are returned.
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is
        an outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a
        value belongs to.


    References
    ----------
    .. [1] Fisch, A., Eckley, I. A., & Fearnhead, P. (2018). A linear time method for
    the detection of point and collective anomalies. arXiv preprint arXiv:1806.01947.
    .. [2] Fisch, A. T., Eckley, I. A., & Fearnhead, P. (2022). Subset multivariate
    collective and point anomaly detection. Journal of Computational and Graphical
    Statistics, 31(2), 574-585.

    Examples
    --------
    from skchange.anomaly_detectors.capa import Capa
    from skchange.datasets.generate import generate_teeth_data

    df = generate_teeth_data(n_segments=5, mean=10, segment_length=100)
    detector = Capa()
    detector.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        saving: Union[str, Tuple[Callable, Callable]] = "mean",
        collective_penalty_scale: float = 2.0,
        point_penalty_scale: float = 2.0,
        min_segment_length: int = 2,
        max_segment_length: int = 1000,
        ignore_point_anomalies: bool = False,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.saving = saving
        self.collective_penalty_scale = collective_penalty_scale
        self.point_penalty_scale = point_penalty_scale
        self.min_segment_length = min_segment_length
        self.max_segment_length = max_segment_length
        self.ignore_point_anomalies = ignore_point_anomalies
        super().__init__(fmt=fmt, labels=labels)

        self.saving_func, self.saving_init_func = saving_factory(self.saving)

        check_larger_than(0, collective_penalty_scale, "collective_penalty_scale")
        check_larger_than(0, point_penalty_scale, "point_penalty_scale")
        check_larger_than(2, min_segment_length, "min_segment_length")
        check_larger_than(min_segment_length, max_segment_length, "max_segment_length")

    def _get_penalty_components(self, X: pd.DataFrame) -> Tuple[np.ndarray, float]:
        # TODO: Add penalty tuning.
        # if self.tune:
        #     return self._tune_threshold(X)
        n = X.shape[0]
        p = X.shape[1]
        # TODO: Add support for depending on 'score'. May interact with p.
        #       E.g. if score is multivariate normal with unknown covariance.
        n_params = 1
        # The default penalty is inflated by a factor of 2 as it is based on Gaussian
        # data. Most data is more heavy-tailed, so we use a bigger penalty.
        # In addition, false positive control is often more important than higher power.
        collective_penalty = dense_capa_penalty(
            n, p, n_params, self.collective_penalty_scale
        )[0]
        point_penalty = self.point_penalty_scale * n_params * p * np.log(n)
        return collective_penalty, point_penalty

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the penalty of the detector.
        If `penalty_scale` is None, the penalty is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `penalty_scale` is a
        number, the penalty is set to `penalty_scale` times the default penalty
        for the detector. The default penalty depends at least on the data's shape,
        but could also depend on more parameters.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self

        State change
        ------------
        creates fitted model (attributes ending in "_")
        """
        X = check_data(
            X,
            min_length=self.min_segment_length,
            min_length_name="min_segment_length",
        )
        self.collective_penalty_, self.point_penalty_ = self._get_penalty_components(X)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        X = check_data(
            X,
            min_length=self.min_segment_length,
            min_length_name="min_segment_length",
        )
        opt_savings, self.collective_anomalies, self.point_anomalies = run_capa(
            X.values,
            self.saving_func,
            self.saving_init_func,
            self.collective_penalty_,
            self.point_penalty_,
            self.min_segment_length,
            self.max_segment_length,
        )
        self.scores = pd.Series(opt_savings, index=X.index, name="score")
        anomalies = format_anomaly_output(
            self.fmt,
            self.labels,
            X.index,
            self.collective_anomalies,
            self.point_anomalies if not self.ignore_point_anomalies else None,
            scores=self.scores,
        )
        return anomalies

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"saving": "mean", "min_segment_length": 5, "max_segment_length": 100},
        ]
        return params



"""Factory for getting saving functions and their initializers from strings.

Savings are cost differences between a baseline cost and an optimised alternative cost.

Recipe for adding new savings:
    1. Add a new module in skchange.costs: <saving_name>_saving.py
    2. Add two functions:
        init_<saving_name>_saving, which precomputes quantities that should be
        precomputed, often partial sums and such.
        <saving_name>_saving, which takes in the output of init_<saving_name>_saving as
        its first argument, and starts and ends indices of the savings to be computed as
        its second and third arguments.
    3. Add the name of the saving to VALID_SAVINGS below.
    4. Add the name of the saving to the docstring of saving_factory below.
    4. Add a new if-statement in saving_factory below.

"""

from skchange.costs.mean_saving import init_mean_saving, mean_saving

VALID_SAVINGS = ["mean"]


def saving_factory(saving_name: str):
    """Return saving function and its initializer.

    Parameters
    ----------
    saving_name : str
        Name of saving function. Must be one of 'mean'.

    Returns
    -------
    cost_func : Callable
        Cost function.
    init_cost_func : Callable
        Cost function initializer.
    """
    if saving_name == "mean":
        return mean_saving, init_mean_saving
    else:
        message = (
            f"saving_name={saving_name} not recognized."
            + f" Must be one of {', '.join(VALID_SAVINGS)}"
        )
        raise ValueError(message)



"""Utilities to profile code."""

import cProfile
import io
import pstats
from pstats import SortKey


class Profiler:
    """Profiler class to profile code."""

    def __init__(self):
        self.pr = cProfile.Profile()
        pass

    def start(self):
        """Start profiling."""
        self.pr.enable()
        return self

    def stop(self):
        """Stop profiling and print sorted results."""
        self.pr.disable()
        s = io.StringIO()
        sortby = SortKey.CUMULATIVE
        ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())  # noqa



import plotly.express as px

from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.mvcapa import Mvcapa
from skchange.datasets.generate import generate_teeth_data
from skchange.utils.benchmarking.profiler import Profiler

# Unviariate
df = generate_teeth_data(n_segments=5, mean=10, segment_length=10, p=1, random_state=2)
capa = Capa(fmt="sparse", max_segment_length=20)
anomalies = capa.fit_predict(df)

capa = Capa(labels="score", fmt="dense", max_segment_length=20)
scores = capa.fit_predict(df)

capa = Capa(labels="indicator", fmt="dense", max_segment_length=20)
anomalies = capa.fit_predict(df)
px.scatter(x=df.index, y=df.values[:, 0], color=anomalies)

# Multivariate
# TODO: Add plotting functionality to assess the affected subset.
df = generate_teeth_data(5, 10, p=10, mean=10, affected_proportion=0.2, random_state=2)
capa = Mvcapa(collective_penalty="sparse", fmt="sparse")
anomalies = capa.fit_predict(df)

capa = Mvcapa(labels="score", fmt="dense", max_segment_length=20)
scores = capa.fit_predict(df)

capa = Mvcapa(collective_penalty_scale=5, labels="indicator", fmt="dense")
anomalies = capa.fit_predict(df)
df.plot(kind="line", backend="plotly")
anomalies.plot(kind="line", backend="plotly")


# Profiling
n = int(1e5)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = Capa(
    max_segment_length=100, collective_penalty_scale=5, point_penalty_scale=5
)
detector = Mvcapa(
    max_segment_length=1000,
    collective_penalty="sparse",
    collective_penalty_scale=5,
    point_penalty_scale=5,
)
profiler = Profiler().start()
detector.fit_predict(df)
profiler.stop()



"""Anomaly detectors composed of change detectors and some conversion logic."""

from typing import Callable, Optional, Union

import numpy as np
import pandas as pd
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.anomaly_detectors.utils import format_anomaly_output


class StatThresholdAnomaliser(BaseSeriesAnnotator):
    """Anomaly detection based on thresholding the values of segment statistics.

    Parameters
    ----------
    change_detector : BaseSeriesAnnotator
        Change detector to use for detecting segments.
    stat : callable, optional (default=np.mean)
        Statistic to calculate per segment.
    stat_lower : float, optional (default=-1.0)
        Segments with a statistic lower than this value are considered anomalous.
    stat_upper : float, optional (default=1.0)
        Segments with a statistic higher than this value are considered anomalous.
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": False,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        change_detector: BaseSeriesAnnotator,
        stat: Callable = np.mean,
        stat_lower: float = -1.0,
        stat_upper: float = 1.0,
        fmt: str = "sparse",
        labels: str = "indicator",
    ):
        self.change_detector = change_detector
        self.stat = stat
        self.stat_lower = stat_lower
        self.stat_upper = stat_upper
        super().__init__(fmt=fmt, labels=labels)

        if self.stat_lower > self.stat_upper:
            message = f"stat_lower ({self.stat_lower}) must be less"
            +f" than or equal to stat_upper ({self.stat_upper})."
            raise ValueError(message)

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fits the change detector to training data.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self
        """
        self.change_detector.fit(X, Y)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        # This is the required output format for the rest of the code to work.
        self.change_detector.fmt = "dense"
        self.change_detector.labels = "int_label"
        self.segments = self.change_detector.predict(X)

        df = pd.concat([X, self.segments], axis=1)
        self.anomalies = []
        for _, segment in df.reset_index(drop=True).groupby("int_label"):
            segment_stat = self.stat(segment.iloc[:, 0].values)
            if (segment_stat < self.stat_lower) | (segment_stat > self.stat_upper):
                self.anomalies.append((int(segment.index[0]), int(segment.index[-1])))

        return format_anomaly_output(
            self.fmt,
            self.labels,
            X.index,
            self.anomalies,
            scores=self.change_detector.scores,
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from skchange.change_detectors.moscore import Moscore

        params = [
            {
                "change_detector": Moscore(bandwidth=10),
                "stat": np.mean,
                "stat_lower": -1.0,
                "stat_upper": 1.0,
            },
        ]
        return params



"""The Moving Score algorithm for multiple collective anomaly detection."""

__author__ = ["mtveten"]
__all__ = ["MoscoreAnomaly"]

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.anomaly_detectors.circular_binseg import greedy_anomaly_selection
from skchange.anomaly_detectors.utils import format_anomaly_output
from skchange.scores.score_factory import anomaly_score_factory
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_larger_than, check_smaller_than


def run_moscore_anomaly(
    X: np.ndarray,
    score_f: Callable,
    score_init_f: Callable,
    anomaly_lengths: np.ndarray,
    left_bandwidth: int,
    right_bandwidth: int,
    threshold: float,
) -> Tuple[List[Tuple[int, int]], np.ndarray, np.ndarray, np.ndarray]:
    n = len(X)
    starts = tuple(
        np.arange(left_bandwidth, n - k - right_bandwidth)
        for k in anomaly_lengths
        if left_bandwidth + k + right_bandwidth <= n
    )
    ends = tuple(
        start + k - 1
        for start, k in zip(starts, anomaly_lengths)
        if left_bandwidth + k + right_bandwidth <= n
    )
    starts = np.concatenate(starts)
    ends = np.concatenate(ends)
    background_starts = starts - left_bandwidth
    background_ends = ends + right_bandwidth
    params = score_init_f(X)
    scores = score_f(params, background_starts, background_ends, starts, ends)
    anomalies = greedy_anomaly_selection(
        scores,
        starts,
        ends,
        background_starts,
        background_ends,
        threshold,
    )
    return anomalies, scores, starts, ends


class MoscoreAnomaly(BaseSeriesAnnotator):
    """Moving score algorithm for multiple collective anomaly detection.

    A custom version of the MOSUM (moving sum) algorithm [1]_ for collective anomaly
    detection. It runs a test statistic for a single anomaly of user-specified lengths
    across all the data, and compares the values in the anomaly window with
    `left_bandwidth` values to the left and `right_bandwidth` samples to the right of
    the anomaly window.

    Experimental for now.

    Efficently implemented using numba.

    Parameters
    ----------
    score: str, Tuple[Callable, Callable], optional (default="mean")
        Test statistic to use for changepoint detection.
        * If "mean", the difference-in-mean statistic is used,
        * If "var", the difference-in-variance statistic is used,
        * If a tuple, it must contain two functions: The first function is the scoring
        function, which takes in the output of the second function as its first
        argument, and start, end and split indices as the second, third and fourth
        arguments. The second function is the initializer, which precomputes quantities
        that should be precomputed. See skchange/scores/score_factory.py for examples.
    min_anomaly_length : int, optional (default=2)
        Minimum length of a collective anomaly.
    max_anomaly_length : int, optional (default=100)
        Maximum length of a collective anomaly. Must be no larger than
        `left_bandwidth + right_bandwidth`.
    left_bandwidth : int, optional (default=50)
        Number of samples to the left of the anomaly window to use in the test
        statistic.
    right_bandwidth : int, optional (default=left_bandwidth)
        Number of samples to the right of the anomaly window to use in the test
        statistic. If None, set to `left_bandwidth`.
    threshold_scale : float, optional (default=2.0)
        Scaling factor for the threshold. The threshold is set to
        'threshold_scale * default_threshold', where the default threshold depends on
        the number of samples, the number of variables, `bandwidth` and `level`.
        If None, the threshold is tuned on the data input to .fit().
    level : float, optional (default=0.01)
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the anomaly score on the training data. For this to be correct, the
        training data must contain no anomalies.
    anomaly_lengths : np.ndarray, optional (default=None)
        Lengths of anomalies to consider. If None, all lengths between
        `min_anomaly_length` and `max_anomaly_length` are considered. If it is not
        important to consider all candidates, just a sparse subset for example,
        customising the anomaly lengths can significantly speed up the algorithm.
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is an
        outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a value
        belongs to.

    References
    ----------
    .. [1] Eichinger, B., & Kirch, C. (2018). A MOSUM procedure for the estimation of
    multiple random change points.

    Examples
    --------
    from skchange.anomaly_detectors.moscore_anomaly import MoscoreAnomaly
    from skchange.datasets.generate import generate_teeth_data

    anomalies = [(100, 119), (250, 299)]
    df = generate_anomalous_data(500, anomalies=anomalies, means=[10.0, 5.0])
    detector = MoscoreAnomaly()
    detector.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        score: Union[str, Tuple[Callable, Callable]] = "mean",
        min_anomaly_length: int = 2,
        max_anomaly_length: int = 100,
        left_bandwidth: int = 50,
        right_bandwidth: int = None,
        threshold_scale: Optional[float] = 2.0,
        level: float = 0.01,
        anomaly_lengths: np.ndarray = None,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.score = score
        self.min_anomaly_length = min_anomaly_length
        self.max_anomaly_length = max_anomaly_length
        self.left_bandwidth = left_bandwidth
        self.right_bandwidth = right_bandwidth
        self.threshold_scale = threshold_scale
        self.level = level
        self.anomaly_lengths = anomaly_lengths
        super().__init__(fmt=fmt, labels=labels)

        self.score_f, self.score_init_f = anomaly_score_factory(score)
        self._right_bandwidth = right_bandwidth if right_bandwidth else left_bandwidth
        if anomaly_lengths is None:
            self._anomaly_lengths = np.arange(
                min_anomaly_length, max_anomaly_length + 1
            )
            self._min_anomaly_length = min_anomaly_length
            self._max_anomaly_length = max_anomaly_length
        else:
            self._anomaly_lengths = np.asarray(anomaly_lengths)
            self._min_anomaly_length = anomaly_lengths.min()
            self._max_anomaly_length = anomaly_lengths.max()

        check_larger_than(1, self._min_anomaly_length, "_min_anomaly_length")
        check_larger_than(
            self._min_anomaly_length, self._max_anomaly_length, "_max_anomaly_length"
        )
        check_smaller_than(
            self.left_bandwidth + self._right_bandwidth,
            self._max_anomaly_length,
            "_max_anomaly_length",
        )
        check_larger_than(1, self.left_bandwidth, "left_bandwidth")
        check_larger_than(1, self._right_bandwidth, "_right_bandwidth")
        check_larger_than(0, threshold_scale, "threshold_scale", allow_none=True)
        check_larger_than(0, self.level, "level")

    def _tune_threshold(self, X: pd.DataFrame) -> float:
        """Tune the threshold for the Moscore algorithm.

        The threshold is set to the (1-`level`)-quantile of the score on the training
        data `X`. For this to be correct, the training data must contain no
        changepoints.

        TODO: Find the threshold given an input number `k` of "permitted" changepoints
        in the training data. This can be achieved by filtering out the top `k` peaks
        of the score.

        Parameters
        ----------
        X : pd.DataFrame
            Training data to tune the threshold on.
        """
        _, scores, _, _ = run_moscore_anomaly(
            X.values,
            self.score_f,
            self.score_init_f,
            self._anomaly_lengths,
            self.left_bandwidth,
            self._right_bandwidth,
            np.inf,
        )
        tuned_threshold = np.quantile(scores, 1 - self.level)
        return tuned_threshold

    @staticmethod
    def get_default_threshold(n: int, p: int) -> float:
        """Get the default threshold for the MoscoreAnomaly algorithm.

        It is the asymptotic critical value of the univariate 'mean' test statitic,
        multiplied by `p` to account for the multivariate case.

        Parameters
        ----------
        n : int
            Sample size.
        p : int
            Number of variables.
        bandwidth : int
            Bandwidth of the Moscore algorithm.
        level : float, optional (default=0.01)
            Significance level for the test statistic.

        Returns
        -------
        threshold : float
            Threshold for the Moscore algorithm.
        """
        return 2 * p * np.sqrt(np.log(n))

    def _get_threshold(self, X: pd.DataFrame) -> float:
        if self.threshold_scale is None:
            return self._tune_threshold(X)
        else:
            n = X.shape[0]
            p = X.shape[1]
            return self.threshold_scale * self.get_default_threshold(n, p)

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the threshold of the detector.
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `threshold_scale` is a
        number, the threshold is set to `threshold_scale` times the default threshold
        for the detector. The default threshold depends at least on the data's shape,
        but could also depend on more parameters.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self
        """
        min_length = (
            self.left_bandwidth + self._right_bandwidth + self._min_anomaly_length
        )
        X = check_data(
            X,
            min_length=min_length,
            min_length_name="left_bandwidth + _right_bandwidth + _min_anomaly_length",
        )
        self.threshold_ = self._get_threshold(X)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        min_length = (
            self.left_bandwidth + self._right_bandwidth + self._min_anomaly_length
        )
        X = check_data(
            X,
            min_length=min_length,
            min_length_name="left_bandwidth + _right_bandwidth + _min_anomaly_length",
        )
        self.anomalies, scores, starts, ends = run_moscore_anomaly(
            X.values,
            self.score_f,
            self.score_init_f,
            self._anomaly_lengths,
            self.left_bandwidth,
            self._right_bandwidth,
            self.threshold_,
        )
        self.scores = pd.DataFrame(
            {"anomaly_start": starts, "anomaly_end": ends, "score": scores}
        )
        return format_anomaly_output(
            self.fmt, self.labels, X.index, self.anomalies, scores=self.scores
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {
                "score": "mean",
                "min_anomaly_length": 5,
                "max_anomaly_length": 100,
                "left_bandwidth": 50,
            },
        ]
        return params



"""Tests for all available changepoint scores."""

import numpy as np
import pytest

from skchange.datasets.generate import generate_teeth_data
from skchange.scores.score_factory import VALID_CHANGE_SCORES, score_factory


@pytest.mark.parametrize("score", VALID_CHANGE_SCORES)
def test_scores(score):
    """Test all available changepoint scores."""
    n = 50
    df = generate_teeth_data(n_segments=1, segment_length=n, p=1, random_state=5)
    score_f, init_score_f = score_factory(score)
    params = init_score_f(df.values)
    scores = np.zeros(n)
    for split in np.arange(10, n - 10).reshape(-1, 1):
        score_value = score_f(
            params, start=np.array([0]), end=np.array([49]), split=split
        )
        assert isinstance(score_value, np.ndarray)
        scores[split] = score_value

    assert np.all(scores >= 0.0)


def test_custom_score():
    """Test custom score."""
    with pytest.raises(ValueError):
        # Need to be jitted to work.
        # Cannot test jitted function because numba is turned of in CI testing.

        def init_score_f(X: np.ndarray) -> np.ndarray:
            return X

        def score_f(params: np.ndarray, start: int, end: int, split: int) -> float:
            return 10.0

        score_factory((score_f, init_score_f))



"""The subset multivariate collective and point anomalies (MVCAPA) algorithm."""

__author__ = ["mtveten"]
__all__ = ["Mvcapa"]

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from scipy.stats import chi2
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.anomaly_detectors.utils import format_multivariate_anomaly_output
from skchange.costs.saving_factory import saving_factory
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_larger_than


def dense_capa_penalty(
    n: int, p: int, n_params: int = 1, scale: float = 1.0
) -> Tuple[float, np.ndarray]:
    """Penalty function for dense anomalies in CAPA.

    Parameters
    ----------
    n : int
        Sample size.
    p : int
        Dimension of the data.
    n_params : int, optional (default=1)
        Number of parameters per segment in the model.
    scale : float, optional (default=1.0)
        Scaling factor for the penalty components.

    Returns
    -------
    alpha : float
        Constant/global penalty term.
    betas : np.ndarray
        Per-component penalty terms.
    """
    psi = np.log(n)
    penalty = scale * (p * n_params + 2 * np.sqrt(p * n_params * psi) + 2 * psi)
    return penalty, np.zeros(p)


def sparse_capa_penalty(
    n: int, p: int, n_params: int = 1, scale: float = 1.0
) -> Tuple[float, np.ndarray]:
    """Penalty function for sparse anomalies in CAPA.

    Parameters
    ----------
    n : int
        Sample size.
    p : int
        Dimension of the data.
    n_params : int, optional (default=1)
        Number of parameters per segment in the model.
    scale : float, optional (default=1.0)
        Scaling factor for the penalty components.

    Returns
    -------
    alpha : float
        Constant/global penalty term.
    betas : np.ndarray
        Per-component penalty terms.
    """
    psi = np.log(n)
    dense_penalty = 2 * scale * psi
    sparse_penalty = 2 * scale * np.log(n_params * p)
    return dense_penalty, np.full(p, sparse_penalty)


def intermediate_capa_penalty(
    n: int, p: int, n_params: int = 1, scale: float = 1.0
) -> Tuple[float, np.ndarray]:
    """Penalty function balancing both dense and sparse anomalies in CAPA.

    Parameters
    ----------
    n : int
        Sample size.
    p : int
        Dimension of the data.
    n_params : int, optional (default=1)
        Number of parameters per segment in the model.
    scale : float, optional (default=1.0)
        Scaling factor for the penalty components.

    Returns
    -------
    alpha : float
        Constant/global penalty term.
    betas : np.ndarray
        Per-component penalty terms.
    """
    if p < 2:
        raise ValueError("p must be at least 2.")

    def penalty_func(j: int) -> float:
        psi = np.log(n)
        c_j = chi2.ppf(1 - j / p, n_params)
        f_j = chi2.pdf(c_j, n_params)
        return scale * (
            2 * (psi + np.log(p))
            + j * n_params
            + 2 * p * c_j * f_j
            + 2 * np.sqrt((j * n_params + 2 * p * c_j * f_j) * (psi + np.log(p)))
        )

    # Penalty function is not defined for j = p.
    penalties = np.vectorize(penalty_func)(np.arange(1, p))
    return 0.0, np.diff(penalties, prepend=0.0, append=penalties[-1])


def combined_capa_penalty(
    n: int, p: int, n_params: int = 1, scale: float = 1.0
) -> Tuple[float, np.ndarray]:
    """Pointwise minimum of dense, sparse and intermediate penalties in CAPA.

    Parameters
    ----------
    n : int
        Sample size.
    p : int
        Dimension of the data.
    n_params : int, optional (default=1)
        Number of parameters per segment in the model.
    scale : float, optional (default=1.0)
        Scaling factor for the penalty components.

    Returns
    -------
    alpha : float
        Constant/global penalty term.
    betas : np.ndarray
        Per-component penalty terms.
    """
    if p < 2:
        return dense_capa_penalty(n, 1, n_params, scale)

    dense_alpha, dense_betas = dense_capa_penalty(n, p, n_params, scale)
    sparse_alpha, sparse_betas = sparse_capa_penalty(n, p, n_params, scale)
    intermediate_alpha, intermediate_betas = intermediate_capa_penalty(
        n, p, n_params, scale
    )
    dense_penalties = dense_alpha + np.cumsum(dense_betas)
    sparse_penalties = sparse_alpha + np.cumsum(sparse_betas)
    intermediate_penalties = intermediate_alpha + np.cumsum(intermediate_betas)
    pointwise_min_penalties = np.zeros(p + 1)
    pointwise_min_penalties[1:] = np.minimum(
        dense_penalties, np.minimum(sparse_penalties, intermediate_penalties)
    )
    return 0.0, np.diff(pointwise_min_penalties)


def capa_penalty_factory(penalty: Union[str, Callable] = "combined") -> Callable:
    """Get a CAPA penalty function.

    Parameters
    ----------
    penalty : str or Callable, optional (default="combined")
        Penalty function to use for CAPA. If a string, must be one of "dense",
        "sparse", "intermediate" or "combined". If a Callable, must be a function
        returning a penalty and per-component penalties, given n, p, n_params and scale.

    Returns
    -------
    penalty_func : Callable
        Penalty function.
    """
    if callable(penalty):
        return penalty
    elif penalty == "dense":
        return dense_capa_penalty
    elif penalty == "sparse":
        return sparse_capa_penalty
    elif penalty == "intermediate":
        return intermediate_capa_penalty
    elif penalty == "combined":
        return combined_capa_penalty
    else:
        raise ValueError(f"Unknown penalty: {penalty}")


@njit
def get_anomalies(
    anomaly_starts: np.ndarray,
) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
    collective_anomalies = []
    point_anomalies = []
    i = anomaly_starts.size - 1
    while i >= 0:
        start_i = anomaly_starts[i]
        size = i - start_i + 1
        if size > 1:
            collective_anomalies.append((int(start_i), i))
            i = int(start_i)
        elif size == 1:
            point_anomalies.append((i, i))
        i -= 1
    return collective_anomalies, point_anomalies


@njit
def penalise_savings(
    savings: np.ndarray, alpha: float, betas: np.ndarray
) -> np.ndarray:
    if np.all(betas < 1e-8):
        penalised_savings = savings.sum(axis=1) - alpha
    if np.all(betas == betas[0]):
        penalised_saving_matrix = np.maximum(savings - betas[0], 0.0) - alpha
        penalised_savings = penalised_saving_matrix.sum(axis=1)
    else:
        n_savings = savings.shape[0]
        penalised_savings = np.zeros(n_savings)
        for i in range(n_savings):
            saving_i = savings[i]
            saving_order = (-saving_i).argsort()  # Decreasing order.
            penalised_saving = np.cumsum(saving_i[saving_order] - betas) - alpha
            argmax = np.argmax(penalised_saving)
            penalised_savings[i] = penalised_saving[argmax]
    return penalised_savings


@njit
def find_affected_components(
    params: Union[np.ndarray, tuple],
    saving_func: Callable,
    anomalies: List[Tuple[int, int]],
    alpha: float,
    betas: np.ndarray,
) -> List[Tuple[int, int, np.ndarray]]:
    new_anomalies = []
    for start, end in anomalies:
        saving = saving_func(params, np.array([start]), np.array([end]))[0]
        saving_order = (-saving).argsort()  # Decreasing order.
        penalised_saving = np.cumsum(saving[saving_order] - betas) - alpha
        argmax = np.argmax(penalised_saving)
        new_anomalies.append((start, end, saving_order[: argmax + 1]))
    return new_anomalies


@njit
def optimise_savings(
    starts: np.ndarray,
    opt_savings: np.ndarray,
    next_savings: np.ndarray,
    alpha: float,
    betas: np.ndarray,
) -> Tuple[float, int, np.ndarray]:
    penalised_saving = penalise_savings(next_savings, alpha, betas)
    candidate_savings = opt_savings[starts] + penalised_saving
    argmax = np.argmax(candidate_savings)
    opt_start = starts[0] + argmax
    return candidate_savings[argmax], opt_start, candidate_savings


@njit
def run_base_capa(
    X: np.ndarray,
    params: Union[np.ndarray, tuple],
    saving_func: Callable,
    collective_alpha: float,
    collective_betas: np.ndarray,
    point_alpha: float,
    point_betas: np.ndarray,
    min_segment_length: int,
    max_segment_length: int,
) -> Tuple[np.ndarray, List[Tuple[int, int]], List[Tuple[int, int]]]:
    n = X.shape[0]
    opt_savings = np.zeros(n + 1)
    # Store the optimal start and affected components of an anomaly for each t.
    # Used to get the final set of anomalies after the loop.
    opt_anomaly_starts = np.repeat(np.nan, n)
    starts = np.array([0])

    ts = np.arange(min_segment_length - 1, n)
    for t in ts:
        # Collective anomalies
        t_array = np.array([t])
        starts = np.concatenate((starts, t_array - min_segment_length + 1))
        ends = np.repeat(t, len(starts))
        collective_savings = saving_func(params, starts, ends)
        opt_collective_saving, opt_start, candidate_savings = optimise_savings(
            starts, opt_savings, collective_savings, collective_alpha, collective_betas
        )

        # Point anomalies
        point_savings = saving_func(params, t_array, t_array)
        opt_point_saving, _, _ = optimise_savings(
            t_array, opt_savings, point_savings, point_alpha, point_betas
        )

        # Combine and store results
        savings = np.array([opt_savings[t], opt_collective_saving, opt_point_saving])
        argmax = np.argmax(savings)
        opt_savings[t + 1] = savings[argmax]
        if argmax == 1:
            opt_anomaly_starts[t] = opt_start
        elif argmax == 2:
            opt_anomaly_starts[t] = t

        # Pruning the admissible starts
        penalty_sum = collective_alpha + collective_betas.sum()
        saving_too_low = candidate_savings + penalty_sum < opt_savings[t + 1]
        too_long_segment = starts < t - max_segment_length + 2
        prune = saving_too_low | too_long_segment
        starts = starts[~prune]

    collective_anomalies, point_anomalies = get_anomalies(opt_anomaly_starts)
    return opt_savings[1:], collective_anomalies, point_anomalies


@njit
def run_mvcapa(
    X: np.ndarray,
    saving_func: Callable,
    saving_init_func: Callable,
    collective_alpha: float,
    collective_betas: np.ndarray,
    point_alpha: float,
    point_betas: np.ndarray,
    min_segment_length: int,
    max_segment_length: int,
) -> Tuple[
    np.ndarray, List[Tuple[int, int, np.ndarray]], List[Tuple[int, int, np.ndarray]]
]:
    params = saving_init_func(X)
    opt_savings, collective_anomalies, point_anomalies = run_base_capa(
        X,
        params,
        saving_func,
        collective_alpha,
        collective_betas,
        point_alpha,
        point_betas,
        min_segment_length,
        max_segment_length,
    )
    collective_anomalies = find_affected_components(
        params,
        saving_func,
        collective_anomalies,
        collective_alpha,
        collective_betas,
    )
    point_anomalies = find_affected_components(
        params, saving_func, point_anomalies, point_alpha, point_betas
    )
    return opt_savings, collective_anomalies, point_anomalies


class Mvcapa(BaseSeriesAnnotator):
    """Subset multivariate collective and point anomaly detection.

    An efficient implementation of the MVCAPA algorithm [1]_ for anomaly detection.

    Parameters
    ----------
    saving : str (default="mean")
        Saving function to use for anomaly detection.
    collective_penalty : str or Callable, optional (default="combined")
        Penalty function to use for collective anomalies. If a string, must be one of
        "dense", "sparse", "intermediate" or "combined". If a Callable, must be a
        function returning a penalty and per-component penalties, given n, p, n_params
        and scale.
    collective_penalty_scale : float, optional (default=1.0)
        Scaling factor for the collective penalty.
    point_penalty : str or Callable, optional (default="sparse")
        Penalty function to use for point anomalies. See 'collective_penalty'.
    point_penalty_scale : float, optional (default=1.0)
        Scaling factor for the point penalty.
    min_segment_length : int, optional (default=2)
        Minimum length of a segment.
    max_segment_length : int, optional (default=1000)
        Maximum length of a segment.
    ignore_point_anomalies : bool, optional (default=False)
        If True, detected point anomalies are not returned by .predict(). I.e., only
        collective anomalies are returned.
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is
        an outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a
        value belongs to.


    References
    ----------
    .. [1] Fisch, A. T., Eckley, I. A., & Fearnhead, P. (2022). Subset multivariate
    collective and point anomaly detection. Journal of Computational and Graphical
    Statistics, 31(2), 574-585.

    Examples
    --------
    from skchange.anomaly_detectors.capa import Capa
    from skchange.datasets.generate import generate_teeth_data

    df = generate_teeth_data(5, 10, p=10, mean=10, affected_proportion=0.2)
    capa = Capa(collective_penalty_scale=5, fmt="sparse", max_segment_length=20)
    capa.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        saving: Union[str, Tuple[Callable, Callable]] = "mean",
        collective_penalty: Union[str, Callable] = "combined",
        collective_penalty_scale: float = 2.0,
        point_penalty: Union[str, Callable] = "sparse",
        point_penalty_scale: float = 2.0,
        min_segment_length: int = 2,
        max_segment_length: int = 1000,
        ignore_point_anomalies: bool = False,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.saving = saving
        self.collective_penalty = collective_penalty
        self.collective_penalty_scale = collective_penalty_scale
        self.point_penalty = point_penalty
        self.point_penalty_scale = point_penalty_scale
        self.min_segment_length = min_segment_length
        self.max_segment_length = max_segment_length
        self.ignore_point_anomalies = ignore_point_anomalies
        super().__init__(fmt=fmt, labels=labels)

        self.saving_func, self.saving_init_func = saving_factory(self.saving)

        check_larger_than(0, collective_penalty_scale, "collective_penalty_scale")
        check_larger_than(0, point_penalty_scale, "point_penalty_scale")
        check_larger_than(2, min_segment_length, "min_segment_length")
        check_larger_than(min_segment_length, max_segment_length, "max_segment_length")

    def _get_penalty_components(self, X: pd.DataFrame) -> Tuple[np.ndarray, float]:
        # TODO: Add penalty tuning.
        # if self.tune:
        #     return self._tune_threshold(X)
        n = X.shape[0]
        p = X.shape[1]
        n_params = 1  # TODO: Add support for depending on 'score'.
        collective_penalty_func = capa_penalty_factory(self.collective_penalty)
        collective_alpha, collective_betas = collective_penalty_func(
            n, p, n_params, scale=self.collective_penalty_scale
        )
        point_penalty_func = capa_penalty_factory(self.point_penalty)
        point_alpha, point_betas = point_penalty_func(
            n, p, n_params, scale=self.point_penalty_scale
        )
        return collective_alpha, collective_betas, point_alpha, point_betas

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the penalty of the detector.
        If `penalty_scale` is None, the penalty is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `penalty_scale` is a
        number, the penalty is set to `penalty_scale` times the default penalty
        for the detector. The default penalty depends at least on the data's shape,
        but could also depend on more parameters.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self

        State change
        ------------
        creates fitted model (attributes ending in "_")
        """
        X = check_data(
            X,
            min_length=self.min_segment_length,
            min_length_name="min_segment_length",
        )
        penalty_components = self._get_penalty_components(X)
        self.collective_alpha_ = penalty_components[0]
        self.collective_betas_ = penalty_components[1]
        self.point_alpha_ = penalty_components[2]
        self.point_betas_ = penalty_components[3]
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series or pd.DataFrame
            Annotations for sequence X, exact format depends on annotation type.
        """
        X = check_data(
            X,
            min_length=self.min_segment_length,
            min_length_name="min_segment_length",
        )
        opt_savings, self.collective_anomalies, self.point_anomalies = run_mvcapa(
            X.values,
            self.saving_func,
            self.saving_init_func,
            self.collective_alpha_,
            self.collective_betas_,
            self.point_alpha_,
            self.point_betas_,
            self.min_segment_length,
            self.max_segment_length,
        )
        self.scores = pd.Series(opt_savings, index=X.index, name="score")
        anomalies = format_multivariate_anomaly_output(
            self.fmt,
            self.labels,
            X.index,
            X.columns,
            self.collective_anomalies,
            self.point_anomalies if not self.ignore_point_anomalies else None,
            self.scores,
        )
        return anomalies

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"saving": "mean", "min_segment_length": 5, "max_segment_length": 100},
        ]
        return params



"""Tests for all annotators/detectors in skchange."""

import pytest
from sktime.tests.test_switch import run_test_for_class
from sktime.utils.estimator_checks import check_estimator

from skchange.anomaly_detectors.tests.test_anomaly_detectors import anomaly_detectors
from skchange.change_detectors.tests.test_change_detectors import change_detectors

all_annotators = anomaly_detectors + change_detectors


@pytest.mark.parametrize("Estimator", all_annotators)
def test_sktime_annotator_compatibility(Estimator):
    """Check compatibility with sktime annotator interface."""
    if not run_test_for_class(Estimator):
        return None

    check_estimator(Estimator, raise_exceptions=True)



"""Test statistic for differences in the mean and/or variance."""

from typing import Tuple

import numpy as np
from numba import njit

from skchange.utils.numba.general import truncate_below
from skchange.utils.numba.stats import col_cumsum


@njit(cache=True)
def init_meanvar_score(X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Precompute sums and squared sums for 'meanvar_score'.

    Parameters
    ----------
    X : np.ndarray
        2D array.

    Returns
    -------
    sums : np.ndarray
        Cumulative sums of X.
    """
    n = X.shape[0]
    p = X.shape[1]
    # 0.0 as first row to make calculations work also for start = 0 in mean_score
    sums = np.zeros((n + 1, p))
    sums[1:] = col_cumsum(X)
    sums2 = np.zeros((n + 1, p))
    sums2[1:] = col_cumsum(X**2)
    return sums, sums2


@njit(cache=True)
def var_from_sums(sums1: np.ndarray, sums2: np.ndarray, i: np.ndarray, j: np.ndarray):
    """Calculate variance from precomputed sums.

    Parameters
    ----------
    sums1 : np.ndarray
        Cumulative sums of X.
    sums2 : np.ndarray
        Cumulative sums of X**2.
    i : np.ndarray
        Start indices in the original data X.
    j : np.ndarray
        End indices in the original data X.

    Returns
    -------
    var : float
        Variance of X[i:j] (inclusive j).
    """
    n = (j - i + 1).reshape(-1, 1)
    sum1 = sums1[j + 1] - sums1[i]  # Indices is moved one step forward
    sum2 = sums2[j + 1] - sums2[i]  # Indices is moved one step forward
    var = sum2 / n - (sum1 / n) ** 2
    return truncate_below(var, 1e-16)  # standard deviation lower bound of 1e-8


@njit(cache=True)
def meanvar_score(
    precomputed_params: np.ndarray,
    start: np.ndarray,
    end: np.ndarray,
    split: np.ndarray,
) -> np.ndarray:
    """Calculate the score for a change in the mean and/or variance.

    Computes the likelihood ratio test for a change in the mean and/or variance of
    i.i.d. Gaussian data.

    Parameters
    ----------
    precomputed_params : np.ndarray
        Precomputed parameters from init_mean_score.
    start : np.ndarray
        Start indices of the intervals to test for a change in the mean.
    end : np.ndarray
        End indices of the intervals to test for a change in the mean.
    split : np.ndarray
        Split indices of the intervals to test for a change in the mean.

    Returns
    -------
    score : np.ndarray
        Scores for a difference in the mean or variance at the given intervals and
        splits.

    Notes
    -----
    To optimize performance, no checks are performed on (start, split, end).
    """
    sums, sums2 = precomputed_params
    before_var = var_from_sums(sums, sums2, start, split)
    after_var = var_from_sums(sums, sums2, split + 1, end)
    full_var = var_from_sums(sums, sums2, start, end)
    before_n = (split - start + 1).reshape(-1, 1)
    after_n = (end - split).reshape(-1, 1)
    before_term = -before_n * np.log(before_var / full_var)
    after_term = -after_n * np.log(after_var / full_var)
    likelihood_ratio = before_term + after_term
    return np.sum(likelihood_ratio, axis=1)


@njit(cache=True)
def baseline_var_from_sums(
    sums1: np.ndarray,
    sums2: np.ndarray,
    interval_start: np.ndarray,
    interval_end: np.ndarray,
    anomaly_start: np.ndarray,
    anomaly_end: np.ndarray,
):
    """Calculate variance from precomputed sums.

    Parameters
    ----------
    sums1 : np.ndarray
        Cumulative sums of X.
    sums2 : np.ndarray
        Cumulative sums of X**2.
    i : np.ndarray
        Start indices in the original data X.
    j : np.ndarray
        End indices in the original data X.

    Returns
    -------
    var : float
        Variance of X[i:j] (inclusive j).
    """
    n = (interval_end - anomaly_end + anomaly_start - interval_start).reshape(-1, 1)
    sum1 = (
        sums1[interval_end + 1]
        - sums1[anomaly_end + 1]
        + sums1[anomaly_start]
        - sums1[interval_start]
    )
    sum2 = (
        sums2[interval_end + 1]
        - sums2[anomaly_end + 1]
        + sums2[anomaly_start]
        - sums2[interval_start]
    )
    var = sum2 / n - (sum1 / n) ** 2
    return truncate_below(var, 1e-16)  # standard deviation lower bound of 1e-8


@njit(cache=True)
def meanvar_anomaly_score(
    precomputed_params: np.ndarray,
    interval_start: np.ndarray,
    interval_end: np.ndarray,
    anomaly_start: np.ndarray,
    anomaly_end: np.ndarray,
) -> np.ndarray:
    """Calculate the score for an anomaly in the mean and/or variance.

    Computes the likelihood ratio test for a change in the mean and/or variance of
    i.i.d. Gaussian data between the anomaly interval and its complement within
    the overall interval.

    The overall and anomalous intervals must satisfy
    `interval_start > anomaly_start <= anomaly_end <= interval_end`.

    Parameters
    ----------
    precomputed_params : np.ndarray
        Precomputed parameters from init_mean_score.
    interval_start : np.ndarray
        Start indices of the intervals to test for an anomaly in.
    interval_end : np.ndarray
        End indices of the intervals to test for an anomaly in.
    anomaly_start : np.ndarray
        Start indices of the anomalies.
    anomaly_end : np.ndarray
        End indices of the anomalies.

    Returns
    -------
    score : float
        Score for an anomaly in the mean and/or variance.

    Notes
    -----
    To optimize performance, no checks are performed on the inputs.
    """
    sums, sums2 = precomputed_params

    baseline_n = interval_end - anomaly_end + anomaly_start - interval_start
    baseline_n = baseline_n.reshape(-1, 1)
    anomaly_n = (anomaly_end - anomaly_start + 1).reshape(-1, 1)

    baseline_var = baseline_var_from_sums(
        sums, sums2, interval_start, interval_end, anomaly_start, anomaly_end
    )
    anomaly_var = var_from_sums(sums, sums2, anomaly_start, anomaly_end)
    full_var = var_from_sums(sums, sums2, interval_start, interval_end)

    baseline_term = -baseline_n * np.log(baseline_var / full_var)
    anomaly_term = -anomaly_n * np.log(anomaly_var / full_var)

    likelihood_ratio = baseline_term + anomaly_term
    return np.sum(likelihood_ratio, axis=1)



"""Validation functions for input data series."""

from typing import Union

import numpy as np
import pandas as pd


def check_data(
    X: Union[pd.DataFrame, pd.Series, np.ndarray],
    min_length: int,
    min_length_name: str = "min_length",
    allow_missing_values: bool = False,
) -> pd.DataFrame:
    """Check if input data is valid.

    Parameters
    ----------
    X : pd.DataFrame, pd.Series
        Input data to check.
    min_length : int
        Minimum number of samples in X.
    min_length_name : str, optional (default="min_length")
        Name of min_length parameter to be shown in the error message.
    allow_missing_values : bool, optional (default=False)
        Whether to allow missing values in X.

    Returns
    -------
    X : pd.DataFrame
        Input data in pd.DataFrame format.
    """
    if isinstance(X, np.ndarray):
        X = pd.DataFrame(X)
    if X.ndim < 2:
        X = X.to_frame()

    if not allow_missing_values and X.isna().any(axis=None):
        raise ValueError(
            f"X cannot contain missing values: X.isna().sum()={X.isna().sum()}."
        )

    n = X.shape[0]
    if n < min_length:
        raise ValueError(
            f"X must have at least {min_length_name}={min_length} samples"
            + f" (X.shape[0]={n})"
        )
    return X



"""Tests for Moscore and all available scores."""

import numpy as np
import pytest

from skchange.change_detectors.moscore import Moscore
from skchange.datasets.generate import generate_teeth_data
from skchange.scores.score_factory import VALID_CHANGE_SCORES


@pytest.mark.parametrize("score", VALID_CHANGE_SCORES)
def test_moscore_changepoint(score):
    """Test Moscore changepoints."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=2
    )
    detector = Moscore(score, fmt="sparse", labels="int_label")
    changepoints = detector.fit_predict(df)
    assert len(changepoints) == n_segments - 1 and changepoints[0] == seg_len - 1


@pytest.mark.parametrize("score", VALID_CHANGE_SCORES)
def test_moscore_scores(score):
    """Test Moscore scores."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=3
    )
    detector = Moscore(score, fmt="dense", labels="score")
    scores = detector.fit_predict(df)
    assert np.all(scores >= 0.0)


@pytest.mark.parametrize("score", VALID_CHANGE_SCORES)
def test_moscore_tuning(score):
    """Test Moscore tuning."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=4
    )
    detector = Moscore(score, threshold_scale=None, fmt="dense", labels="indicator")
    detector.fit(df)
    assert detector.threshold_ > 0.0



"""Factory for getting test statistic functions and their initializers from strings.

Recipe for adding new scores:
    1. Add a new module in skchange.score: <score_name>_score.py
    2. Add two functions:
        init_<score_name>_score, which precomputes quantities that should be
        precomputed, often partial sums and such.
        <score_name>_score, which takes in the output of init_<score_name>_score as its
        first argument, and start, end and split indices as the second, third and
        fourth arguments.
    3. Add the name of the score to VALID_CHANGE_SCORES below.
    4. Add the name of the score to the docstring of score_factory below.
    4. Add a new if-statement in score_factory below.

"""

from typing import Callable, Tuple, Union

from numba.extending import is_jitted

from skchange.scores.mean_score import init_mean_score, mean_anomaly_score, mean_score
from skchange.scores.meanvar_score import (
    init_meanvar_score,
    meanvar_anomaly_score,
    meanvar_score,
)

VALID_CHANGE_SCORES = ["mean", "meanvar"]
VALID_ANOMALY_SCORES = ["mean", "meanvar"]


def score_factory(score: Union[str, Tuple[Callable, Callable]]):
    """Return score function and its initializer.

    Parameters
    ----------
    score: str, Tuple[Callable, Callable]
        Test statistic to use for changepoint detection.
        * If "mean", the difference-in-mean statistic is used,
        * If "meanvar", the difference-in-mean-and-variance statistic is used,
        * If a tuple, it must contain two functions: The first function is the scoring
        function, which takes in the output of the second function as its first
        argument, and start, end and split indices as the second, third and fourth
        arguments. The second function is the initializer, which precomputes quantities
        that should be precomputed. See skchange/scores/score_factory.py for examples.

    Returns
    -------
    score_func : Numba jitted Callable
        Score function.
    init_score_func : Numba jitted Callable
        Score function initializer.
    """
    if isinstance(score, str) and score == "mean":
        return mean_score, init_mean_score
    elif isinstance(score, str) and score == "meanvar":
        return meanvar_score, init_meanvar_score
    elif len(score) == 2 and all([is_jitted(s) for s in score]):
        return score[0], score[1]
    else:
        message = (
            f"score={score} not recognized."
            + f" Must be one of {', '.join(VALID_CHANGE_SCORES)}"
            + " or a tuple of two numba jitted functions."
        )
        raise ValueError(message)


def anomaly_score_factory(score: Union[str, Tuple[Callable, Callable]]):
    """Return anomaly score function and its initializer.

    Parameters
    ----------
    score: str, Tuple[Callable, Callable]
        Test statistic to use for anomaly detection.
        * If "mean", the difference-in-mean statistic is used,
        * If a tuple, it must contain two functions: The first function is the scoring
        function, which takes in the output of the second function as its first
        argument, and start, end and split indices as the second, third and fourth
        arguments. The second function is the initializer, which precomputes quantities
        that should be precomputed. See skchange/scores/score_factory.py for examples.

    Returns
    -------
    score_func : Numba jitted Callable
        Score function.
    init_score_func : Numba jitted Callable
        Score function initializer.
    """
    if isinstance(score, str) and score == "mean":
        return mean_anomaly_score, init_mean_score
    elif isinstance(score, str) and score == "meanvar":
        return meanvar_anomaly_score, init_meanvar_score
    elif len(score) == 2 and all([is_jitted(s) for s in score]):
        return score[0], score[1]
    else:
        message = (
            f"score={score} not recognized."
            + f" Must be one of {', '.join(VALID_ANOMALY_SCORES)}"
            + " or a tuple of two numba jitted functions."
        )
        raise ValueError(message)



"""Tests for Moscore and all available scores."""

import numpy as np
import pandas as pd
import pytest

from skchange.change_detectors.seeded_binseg import SeededBinarySegmentation
from skchange.datasets.generate import generate_teeth_data
from skchange.scores.score_factory import VALID_CHANGE_SCORES


def test_invalid_parameters():
    """Test invalid input parameters to SeededBinarySegmentation.

    These tests serve as tests for the input validators.
    """
    with pytest.raises(ValueError):
        SeededBinarySegmentation(threshold_scale=-0.1)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(min_segment_length=0)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(min_segment_length=None)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(min_segment_length=5, max_interval_length=9)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(growth_factor=1.0)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(growth_factor=None)
    with pytest.raises(ValueError):
        SeededBinarySegmentation(growth_factor=2.01)


def test_invalid_data():
    """Test invalid input data to SeededBinarySegmentation.

    These tests serve as tests for the input validators.
    """
    detector = SeededBinarySegmentation()
    with pytest.raises(ValueError):
        detector.fit(np.array([1.0]))

    with pytest.raises(ValueError):
        detector.fit(pd.Series([1.0, np.nan, 1.0, 1.0]))


@pytest.mark.parametrize("score", VALID_CHANGE_SCORES)
def test_binseg_tuning(score):
    """Test SeededBinarySegmentation tuning."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=4
    )
    detector = SeededBinarySegmentation(
        score, threshold_scale=None, fmt="dense", labels="score"
    )
    detector.fit(df)
    scores = detector.predict(df)
    assert detector.threshold_ >= scores["score"].mean()
    assert detector.threshold_ <= scores["score"].max()



"""Circular binary segmentation algorithm for multiple changepoint detection."""

__author__ = ["mtveten"]
__all__ = ["CircularBinarySegmentation"]

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.anomaly_detectors.utils import format_anomaly_output
from skchange.change_detectors.seeded_binseg import make_seeded_intervals
from skchange.scores.score_factory import anomaly_score_factory
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_in_interval, check_larger_than


@njit
def greedy_anomaly_selection(
    scores: np.ndarray,
    anomaly_starts: np.ndarray,
    anomaly_ends: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    threshold: float,
) -> List[Tuple[int, int]]:
    scores = scores.copy()
    anomalies = []
    while np.any(scores > threshold):
        argmax = scores.argmax()
        anomaly_start = anomaly_starts[argmax]
        anomaly_end = anomaly_ends[argmax]
        anomalies.append((anomaly_start, anomaly_end))
        scores[(anomaly_end >= starts) & (anomaly_start < ends)] = 0.0
    anomalies.sort()
    return anomalies


@njit
def make_anomaly_intervals(
    interval_start: int, interval_end: int, min_segment_length: int = 1
) -> Tuple[np.ndarray, np.ndarray]:
    starts = []
    ends = []
    for i in range(interval_start + 1, interval_end - min_segment_length + 2):
        for j in range(i + min_segment_length - 1, interval_end + 1):
            baseline_n = interval_end - j + i - interval_start
            if baseline_n >= min_segment_length:
                starts.append(i)
                ends.append(j)
    return np.array(starts, dtype=np.int64), np.array(ends, dtype=np.int64)


@njit
def run_circular_binseg(
    X: np.ndarray,
    score_func: Callable,
    score_init_func: Callable,
    threshold: float,
    min_segment_length: int,
    max_interval_length: int,
    growth_factor: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    starts, ends = make_seeded_intervals(
        X.shape[0],
        2 * min_segment_length,
        max_interval_length,
        growth_factor,
    )
    params = score_init_func(X)

    anomaly_scores = np.zeros(starts.size)
    anomaly_starts = np.zeros(starts.size, dtype=np.int64)
    anomaly_ends = np.zeros(starts.size, dtype=np.int64)
    maximizers = np.zeros((starts.size, 2))
    for i, (start, end) in enumerate(zip(starts, ends)):
        anomaly_start_candidates, anomaly_end_candidates = make_anomaly_intervals(
            start, end, min_segment_length
        )
        scores = score_func(
            params,
            np.repeat(start, anomaly_start_candidates.size),
            np.repeat(end, anomaly_start_candidates.size),
            anomaly_start_candidates,
            anomaly_end_candidates,
        )
        argmax = np.argmax(scores)
        anomaly_scores[i] = scores[argmax]
        anomaly_starts[i] = anomaly_start_candidates[argmax]
        anomaly_ends[i] = anomaly_end_candidates[argmax]

    anomalies = greedy_anomaly_selection(
        anomaly_scores, anomaly_starts, anomaly_ends, starts, ends, threshold
    )
    return anomalies, anomaly_scores, maximizers, starts, ends


class CircularBinarySegmentation(BaseSeriesAnnotator):
    """Circular binary segmentation algorithm for multiple collective anomaly detection.

    Binary segmentation type changepoint detection algorithms recursively split the data
    into two segments, and test whether the two segments are different. Circular binary
    segmentation is described in [1]_.

    Efficently implemented using numba.

    Parameters
    ----------
    score: str, Tuple[Callable, Callable], optional (default="mean")
        Test statistic to use for changepoint detection.
        * If "mean", the difference-in-mean statistic is used,
        * If "var", the difference-in-variance statistic is used,
        * If a tuple, it must contain two functions: The first function is the scoring
        function, which takes in the output of the second function as its first
        argument, and start, end and split indices as the second, third and fourth
        arguments. The second function is the initializer, which precomputes quantities
        that should be precomputed. See skchange/scores/score_factory.py for examples.
    threshold_scale : float, optional (default=2.0)
        Scaling factor for the threshold. The threshold is set to
        'threshold_scale * 2 * p * np.sqrt(np.log(n))', where 'n' is the sample size
        and 'p' is the number of variables. If None, the threshold is tuned on the data
        input to .fit().
    level : float, optional (default=0.01)
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the changepoint scores of all the seeded intervals on the training data.
        For this to be correct, the training data must contain no changepoints.
    min_segment_length : int, optional (default=5)
        Minimum length between two changepoints. Must be greater than or equal to 1.
    max_interval_length : int (default=100)
        The maximum length of an interval to estimate a changepoint in. Must be greater
        than or equal to '2 * min_segment_length'.
    growth_factor : float (default = 1.5)
        The growth factor for the seeded intervals. Intervals grow in size according to
        'interval_len=max(interval_len + 1, np.floor(growth_factor * interval_len))',
        starting at 'interval_len'='min_interval_length'. It also governs the amount
        of overlap between intervals of the same length, as the start of each interval
        is shifted by a factor of '1 + 1 / growth_factor'. Must be a float in (1, 2].
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is an
        outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a value
        belongs to.

    References
    ----------
    .. [1] Olshen, A. B., Venkatraman, E. S., Lucito, R., & Wigler, M. (2004). Circular
    binary segmentation for the analysis of array‐based DNA copy number data.
    Biostatistics, 5(4), 557-572.

    Examples
    --------
    from skchange.anomaly_detectors.circular_binseg import CircularBinarySegmentation
    from skchange.datasets.generate import generate_teeth_data

    # Generate data
    df = generate_teeth_data(
        n_segments=3, mean=10, segment_length=100000, p=5, random_state=2
    )

    # Detect changepoints
    detector = CircularBinarySegmentation()
    detector.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        score: Union[str, Tuple[Callable, Callable]] = "mean",
        threshold_scale: Optional[float] = 2.0,
        level: float = 1e-8,
        min_segment_length: int = 5,
        max_interval_length: int = 100,
        growth_factor: float = 1.5,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.score = score
        self.threshold_scale = threshold_scale  # Just holds the input value.
        self.level = level
        self.min_segment_length = min_segment_length
        self.max_interval_length = max_interval_length
        self.growth_factor = growth_factor
        super().__init__(fmt=fmt, labels=labels)
        self.score_f, self.score_init_f = anomaly_score_factory(self.score)

        check_larger_than(0.0, self.threshold_scale, "threshold_scale", allow_none=True)
        check_in_interval(pd.Interval(0.0, 1.0, closed="neither"), self.level, "level")
        check_larger_than(1.0, self.min_segment_length, "min_segment_length")
        check_larger_than(
            2 * self.min_segment_length, self.max_interval_length, "max_interval_length"
        )
        check_in_interval(
            pd.Interval(1.0, 2.0, closed="right"),
            self.growth_factor,
            "growth_factor",
        )

    def _tune_threshold(self, X: pd.DataFrame) -> float:
        """Tune the threshold.

        The threshold is set to the (1-`level`)-quantile of the changepoint scores from
        all the seeded intervals on the training data `X`. For this to be correct, the
        training data must contain no changepoints.

        Parameters
        ----------
        X : pd.DataFrame
            Training data to tune the threshold on.

        Returns
        -------
        threshold : float
            The tuned threshold.
        """
        _, scores, _, _, _ = run_circular_binseg(
            X.values,
            self.score_f,
            self.score_init_f,
            np.inf,
            self.min_segment_length,
            self.max_interval_length,
            self.growth_factor,
        )
        return np.quantile(scores, 1 - self.level)

    @staticmethod
    def get_default_threshold(n: int, p: int) -> float:
        """Get the default threshold for Circular Binary Segmentation.

        Parameters
        ----------
        n : int
            Sample size.
        p : int
            Number of variables.

        Returns
        -------
        threshold : float
            The default threshold.
        """
        return 2 * p * np.sqrt(np.log(n))

    def _get_threshold(self, X: pd.DataFrame) -> float:
        if self.threshold_scale is None:
            return self._tune_threshold(X)
        else:
            n = X.shape[0]
            p = X.shape[1]
            return self.threshold_scale * self.get_default_threshold(n, p)

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the threshold of the detector.
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `threshold_scale` is a
        number, the threshold is set to `threshold_scale` times the default threshold
        for the detector. The default threshold depends at least on the data's shape,
        but could also depend on more parameters.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self
        """
        X = check_data(
            X,
            min_length=2 * self.min_segment_length,
            min_length_name="min_interval_length",
        )
        self.threshold_ = self._get_threshold(X)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        X = check_data(
            X,
            min_length=2 * self.min_segment_length,
            min_length_name="min_interval_length",
        )
        anomalies, scores, maximizers, starts, ends = run_circular_binseg(
            X.values,
            self.score_f,
            self.score_init_f,
            self.threshold_,
            self.min_segment_length,
            self.max_interval_length,
            self.growth_factor,
        )
        self.anomalies = anomalies
        self.scores = pd.DataFrame(
            {
                "interval_start": starts,
                "interval_end": ends,
                "argmax_anomaly_start": maximizers[:, 0],
                "argmax_anomaly_end": maximizers[:, 1],
                "score": scores,
            }
        )
        return format_anomaly_output(
            self.fmt, self.labels, X.index, self.anomalies, scores=self.scores
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"score": "mean", "min_segment_length": 5, "max_interval_length": 100},
        ]
        return params



"""Utility functions for change detection."""

from typing import Union

import numpy as np
import pandas as pd


def changepoints_to_labels(changepoints: list, n) -> np.ndarray:
    """Convert a list of changepoints to a list of labels.

    Parameters
    ----------
    changepoints : list
        List of changepoint indices.
    n: int
        Sample size.

    Returns
    -------
    labels : np.ndarray
        1D array of labels: 0 for the first segment, 1 for the second, etc.
    """
    changepoints = [-1] + changepoints + [n - 1]
    labels = np.zeros(n)
    for i in range(len(changepoints) - 1):
        labels[changepoints[i] + 1 : changepoints[i + 1] + 1] = i
    return labels


def format_changepoint_output(
    fmt: str,
    labels: str,
    changepoints: list,
    X_index: pd.Index,
    scores: Union[pd.Series, pd.DataFrame] = None,
) -> pd.Series:
    """Format the predict method output of change detectors.

    Parameters
    ----------
    fmt : str
        Format of the output. Either "sparse" or "dense".
    labels : str
        Labels of the output. Either "indicator", "score" or "int_label".
    changepoints : list
        List of changepoint indices.
    X_index : pd.Index
        Index of the input data.
    scores : pd.Series or pd.DataFrame, optional (default=None)
        Series or DataFrame of scores. If Series, it must be named 'score', and if
        DataFrame, it must have a column named 'score'.

    Returns
    -------
    pd.Series
        Either a sparse or dense pd.Series of boolean labels, integer labels or scores.
    """
    if fmt == "sparse" and labels in ["int_label", "indicator"]:
        out = pd.Series(changepoints, name="changepoints", dtype=int)
    elif fmt == "dense" and labels == "int_label":
        out = changepoints_to_labels(changepoints, len(X_index))
        out = pd.Series(out, index=X_index, name="int_label", dtype=int)
    elif fmt == "dense" and labels == "indicator":
        out = pd.Series(False, index=X_index, name="indicator", dtype=bool)
        out.iloc[changepoints] = True
    elif labels == "score":
        # There is no sparse version of 'score'.
        # The scores are formatted in each class' _predict method, as what is a good
        # format for the scores is method dependent.
        out = scores
    return out



import plotly.express as px

from skchange.anomaly_detectors.circular_binseg import (
    CircularBinarySegmentation,
    make_anomaly_intervals,
)
from skchange.datasets.generate import generate_teeth_data
from skchange.utils.benchmarking.profiler import Profiler

df = generate_teeth_data(n_segments=3, mean=10, segment_length=20, p=1, random_state=7)
detector = CircularBinarySegmentation(
    score="mean", growth_factor=1.5, min_segment_length=10
)
anomalies = detector.fit_predict(df)

df.plot(kind="line", backend="plotly")

px.scatter(detector.scores, x="argmax_anomaly_start", y="score")

# Test anomaly intervals
anomaly_intervals = make_anomaly_intervals(0, 5, 2)

# Profiling
n = int(1e5)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = CircularBinarySegmentation("mean", growth_factor=2)
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()



"""Common validation functions for input parameters."""

from numbers import Number

import pandas as pd


def check_larger_than(
    min_value: Number, value: Number, name: str, allow_none: bool = False
) -> Number:
    """Check if value is non-negative.

    Parameters
    ----------
    min_value : int, float
        Minimum allowed value.
    value : int, float
        Value to check.
    name : str
        Name of the parameter to be shown in the error message.
    allow_none : bool, optional (default=False)
        Whether to allow None values.

    Returns
    -------
    value : int, float
        Input value.

    Raises
    ------
    ValueError
        If value is negative.
    """
    if not allow_none and value is None:
        raise ValueError(f"{name} cannot be None.")
    if value is not None and value < min_value:
        raise ValueError(f"{name} must be at least {min_value} ({name}={value}).")
    return value


def check_smaller_than(
    max_value: Number, value: Number, name: str, allow_none: bool = False
) -> Number:
    """Check if value is non-negative.

    Parameters
    ----------
    max_value : int, float
        Maximum allowed value.
    value : int, float
        Value to check.
    name : str
        Name of the parameter to be shown in the error message.
    allow_none : bool, optional (default=False)
        Whether to allow None values.

    Returns
    -------
    value : int, float
        Input value.

    Raises
    ------
    ValueError
        If value is negative.
    """
    if not allow_none and value is None:
        raise ValueError(f"{name} cannot be None.")
    if value is not None and value > max_value:
        raise ValueError(f"{name} must be no larger than {max_value} ({name}={value}).")
    return value


def check_in_interval(
    interval: pd.Interval,
    value: Number,
    name: str,
    allow_none: bool = False,
) -> Number:
    """Check if value is non-negative.

    Parameters
    ----------
    interval : pd.Interval
        Interval to check.
    value : int, float
        Value to check.
    name : str
        Name of the parameter to be shown in the error message.
    allow_none : bool, optional (default=False)
        Whether to allow None values.

    Returns
    -------
    value : int, float
        Input value.

    Raises
    ------
    ValueError
        If value is negative.
    """
    if not allow_none and value is None:
        raise ValueError(f"{name} cannot be None.")
    if value is not None and value not in interval:
        raise ValueError(f"{name} must be in {interval} ({name}={value}).")
    return value



"""Tests for CAPA and all available savings."""

import pytest

from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.mvcapa import Mvcapa
from skchange.costs.saving_factory import VALID_SAVINGS
from skchange.datasets.generate import generate_teeth_data


@pytest.mark.parametrize("saving", VALID_SAVINGS)
def test_capa_anomalies(saving):
    """Test Capa anomalies."""
    n_segments = 2
    seg_len = 20
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=5, random_state=8
    )
    for detector_class in [Capa, Mvcapa]:
        detector = detector_class(
            saving=saving, fmt="sparse", collective_penalty_scale=2.0
        )
        anomalies = detector.fit_predict(df)
        # End point also included as a changepoint
        assert (
            len(anomalies) == 1
            and anomalies.loc[0, "start"] == seg_len
            and anomalies.loc[0, "end"] == 2 * seg_len - 1
        )



import numpy as np
import plotly.express as px
from numba import njit

from skchange.change_detectors.moscore import Moscore, where
from skchange.datasets.generate import add_linspace_outliers, generate_teeth_data
from skchange.scores.mean_score import init_mean_score, mean_score
from skchange.utils.benchmarking.profiler import Profiler

# Simple univariate example
df = generate_teeth_data(n_segments=2, mean=10, segment_length=100, p=1, random_state=2)
detector = Moscore()
changepoints = detector.fit_predict(df)
px.scatter(detector.scores)


# Profiling
n = int(1e6)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = Moscore(bandwidth=50)
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()


# Various unit tests
df = generate_teeth_data(n_segments=1, mean=10, segment_length=10, p=1)
precomputed_params = init_mean_score(df.values)
mean_score(precomputed_params, start=0, end=9, split=4)
where(np.array([True, True, True, False, False]))


# Variance score
df = generate_teeth_data(
    n_segments=2, variance=16, segment_length=100, p=1, random_state=1
)
detector = Moscore(score="meanvar")
changepoints = detector.fit_predict(df)
px.scatter(df)
px.scatter(detector.scores)


# Custom score
@njit
def col_median(X: np.ndarray) -> np.ndarray:
    m = X.shape[1]
    medians = np.zeros(m)
    for j in range(m):
        medians[j] = np.median(X[:, j])
    return medians


@njit
def init_spike_score(X: np.ndarray) -> np.ndarray:
    return X


def spike_score_factory(margin: int = 0):
    @njit
    def spike_score(
        precomputed_params: np.ndarray, start: int, end: int, split: int
    ) -> float:
        X = precomputed_params
        interval_X = np.concatenate(
            (X[start : split - margin], X[split + margin + 1 : end + 1])
        )
        baseline_median = col_median(interval_X)
        return np.sum(np.abs(X[split] - baseline_median))

    return spike_score


df = generate_teeth_data(n_segments=1, mean=0, segment_length=100, p=1)
df = add_linspace_outliers(df, n_outliers=4, outlier_size=10)
score = (spike_score_factory(margin=0), init_spike_score)
detector = Moscore(score, bandwidth=5)
anomalies = detector.fit_predict(df)
px.scatter(detector.scores)
px.scatter(df)



"""Numba-optimized functions for various common data manipulation tasks.""" ""

import numpy as np
from numba import njit, prange


@njit(cache=True)
def col_repeat(x: np.ndarray, n: int) -> np.ndarray:
    """Repeat each column of a 2D array n times.

    Parameters
    ----------
    x : np.ndarray
        1D array.

    Returns
    -------
    2D array : (x.size, n)-matrix with x in each column
    """
    expanded_x = np.zeros((x.shape[0], n))
    for j in prange(n):
        expanded_x[:, j] = x
    return expanded_x


@njit(cache=True)
def row_repeat(x: np.ndarray, n: int) -> np.ndarray:
    """Repeat each row of a 2D array n times.

    Parameters
    ----------
    x : np.ndarray
        1D array.

    Returns
    -------
    2D array : (n, x.size) matrix with x in each row
    """
    expanded_x = np.zeros((x.shape[0], n))
    for i in prange(n):
        expanded_x[i, :] = x
    return expanded_x


@njit(cache=True)
def where(indicator: np.ndarray) -> list:
    """
    Identify consecutive intervals of True values in the input array.

    Parameters
    ----------
    indicator : np.ndarray
        1D boolean array.

    Returns
    -------
    list of tuples:
        Each tuple represents the start and end indices of consecutive True intervals.
        If there are no True values, an empty list is returned.
    """
    intervals = []
    start, end = None, None
    for i, val in enumerate(indicator):
        if val and start is None:
            start = i
        elif not val and start is not None:
            end = i - 1
            intervals.append((start, end))
            start, end = None, None
    if start is not None:
        intervals.append((start, len(indicator) - 1))
    return intervals


@njit(cache=True)
def truncate_below(x: np.ndarray, lower_bound: float) -> np.ndarray:
    """Truncate values below a lower bound.

    Parameters
    ----------
    x : np.ndarray
        2D array.
    lower_bound : float
        Lower bound.

    Returns
    -------
    x : np.ndarray
        2D array with values below lower_bound replaced by lower_bound.
    """
    if x.ndim == 1:
        x[x < lower_bound] = lower_bound
    else:
        p = x.shape[1]
        for j in range(p):
            # Numba doesn't support multidimensional index.
            x[x[:, j] < lower_bound, j] = lower_bound  # To avoid zero division
    return x



"""Basic tests for all anomaly detectors."""

import pandas as pd
import pytest
from sktime.tests.test_switch import run_test_for_class
from sktime.utils._testing.annotation import make_annotation_problem

from skchange.anomaly_detectors.anomalisers import StatThresholdAnomaliser
from skchange.anomaly_detectors.capa import Capa
from skchange.anomaly_detectors.circular_binseg import CircularBinarySegmentation
from skchange.anomaly_detectors.moscore_anomaly import MoscoreAnomaly
from skchange.anomaly_detectors.mvcapa import Mvcapa
from skchange.datasets.generate import generate_anomalous_data

anomaly_detectors = [
    Capa,
    CircularBinarySegmentation,
    MoscoreAnomaly,
    Mvcapa,
    StatThresholdAnomaliser,
]

true_anomalies = [(50, 59), (120, 129)]
anomaly_data = generate_anomalous_data(
    200, anomalies=true_anomalies, means=[10.0, 5.0], random_state=39
)


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_output_type(Estimator):
    """Test annotator output type."""
    estimator = Estimator.create_test_instance()
    if not run_test_for_class(Estimator):
        return None

    arg = make_annotation_problem(
        n_timepoints=500, estimator_type=estimator.get_tag("distribution_type")
    )
    estimator.fit(arg)
    arg = make_annotation_problem(
        n_timepoints=200, estimator_type=estimator.get_tag("distribution_type")
    )
    y_pred = estimator.predict(arg)
    assert isinstance(y_pred, (pd.DataFrame, pd.Series))


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_anomaly_detector_sparse_int(Estimator):
    """Test sparse int label anomaly detector output.

    Check if the predicted anomalies match.
    """
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="sparse", labels="int_label")
    anomalies = detector.fit_predict(anomaly_data)
    assert len(anomalies) == len(true_anomalies)
    for i, (start, end) in enumerate(true_anomalies):
        assert anomalies.loc[i, "start"] == start and anomalies.loc[i, "end"] == end


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_anomaly_detector_sparse_indicator(Estimator):
    """Test sparse indicator anomaly detector output.

    Check if the predicted anomalies match.
    """
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="sparse", labels="indicator")
    anomalies = detector.fit_predict(anomaly_data)
    assert len(anomalies) == len(true_anomalies)
    for i, (start, end) in enumerate(true_anomalies):
        assert anomalies.loc[i, "start"] == start and anomalies.loc[i, "end"] == end


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_anomaly_detector_score(Estimator):
    """Test score anomaly detector output."""
    sparse_detector = Estimator.create_test_instance()
    sparse_detector.set_params(fmt="sparse", labels="score")
    dense_detector = Estimator.create_test_instance()
    dense_detector.set_params(fmt="dense", labels="score")
    sparse_scores = sparse_detector.fit_predict(anomaly_data)
    dense_scores = dense_detector.fit_predict(anomaly_data)
    assert (sparse_scores == dense_scores).all(axis=None)
    if isinstance(sparse_scores, pd.DataFrame):
        assert "score" in sparse_scores.columns
    else:
        assert sparse_scores.name == "score"


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_anomaly_detector_dense_int(Estimator):
    """Tests dense int label anomaly detector output.

    Check if the predicted anomalies matches.
    """
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="dense", labels="int_label")
    labels = detector.fit_predict(anomaly_data)
    if isinstance(labels, pd.DataFrame):
        labels = labels.iloc[:, 0]

    assert labels.nunique() == len(true_anomalies) + 1
    for i, (start, end) in enumerate(true_anomalies):
        assert (labels.iloc[start : end + 1] == i + 1).all()


@pytest.mark.parametrize("Estimator", anomaly_detectors)
def test_anomaly_detector_dense_indicator(Estimator):
    """Tests dense indicator anomaly detector output.

    Check if the predicted anomalies matches.
    """
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="dense", labels="indicator")
    labels = detector.fit_predict(anomaly_data)
    if isinstance(labels, pd.DataFrame):
        labels = labels.iloc[:, 0]

    for start, end in true_anomalies:
        assert labels.iloc[start : end + 1].all()
        assert not labels.iloc[start - 1] and not labels.iloc[end + 1]



"""Basic tests for all change detectors."""

import pandas as pd
import pytest
from sktime.tests.test_switch import run_test_for_class
from sktime.utils._testing.annotation import make_annotation_problem

from skchange.change_detectors.moscore import Moscore
from skchange.change_detectors.pelt import Pelt
from skchange.change_detectors.seeded_binseg import SeededBinarySegmentation
from skchange.datasets.generate import generate_teeth_data

change_detectors = [Moscore, Pelt, SeededBinarySegmentation]


@pytest.mark.parametrize("Estimator", change_detectors)
def test_output_type(Estimator):
    """Test annotator output type."""
    estimator = Estimator.create_test_instance()
    if not run_test_for_class(Estimator):
        return None

    arg = make_annotation_problem(
        n_timepoints=50, estimator_type=estimator.get_tag("distribution_type")
    )
    estimator.fit(arg)
    arg = make_annotation_problem(
        n_timepoints=30, estimator_type=estimator.get_tag("distribution_type")
    )
    y_pred = estimator.predict(arg)
    assert isinstance(y_pred, (pd.DataFrame, pd.Series))


@pytest.mark.parametrize("Estimator", change_detectors)
def test_change_detector_sparse_int(Estimator):
    """Test sparse int_label segmentation."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=2
    )
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="sparse", labels="int_label")
    changepoints = detector.fit_predict(df)
    assert len(changepoints) == n_segments - 1 and changepoints[0] == seg_len - 1


@pytest.mark.parametrize("Estimator", change_detectors)
def test_change_detector_sparse_indicator(Estimator):
    """Test sparse indicator segmentation."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=3
    )
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="sparse", labels="indicator")
    changepoints = detector.fit_predict(df)
    assert len(changepoints) == n_segments - 1 and changepoints[0] == seg_len - 1


@pytest.mark.parametrize("Estimator", change_detectors)
def test_change_detector_score(Estimator):
    """Test sparse score segmentation."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=4
    )
    sparse_detector = Estimator.create_test_instance()
    sparse_detector.set_params(fmt="sparse", labels="score")
    dense_detector = Estimator.create_test_instance()
    dense_detector.set_params(fmt="dense", labels="score")
    sparse_scores = sparse_detector.fit_predict(df)
    dense_scores = dense_detector.fit_predict(df)
    assert (sparse_scores == dense_scores).all(axis=None)
    if isinstance(sparse_scores, pd.DataFrame):
        assert "score" in sparse_scores.columns
    else:
        assert sparse_scores.name == "score"


@pytest.mark.parametrize("Estimator", change_detectors)
def test_change_detector_dense_int(Estimator):
    """Tests dense int_label segmentation."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=2
    )
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="dense", labels="int_label")
    labels = detector.fit_predict(df)
    assert labels.nunique() == n_segments
    assert labels[seg_len - 1] == 0.0 and labels[seg_len] == 1.0


@pytest.mark.parametrize("Estimator", change_detectors)
def test_change_detector_dense_indicator(Estimator):
    """Tests dense indicator segmentation."""
    n_segments = 2
    seg_len = 50
    df = generate_teeth_data(
        n_segments=n_segments, mean=10, segment_length=seg_len, p=1, random_state=8
    )
    detector = Estimator.create_test_instance()
    detector.set_params(fmt="dense", labels="indicator")
    cpt_indicator = detector.fit_predict(df)
    assert cpt_indicator.sum() == n_segments - 1
    assert cpt_indicator[seg_len - 1]



"""The pruned exact linear time (PELT) algorithm."""

__author__ = ["mtveten"]
__all__ = ["Pelt"]


from typing import Callable, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.change_detectors.utils import format_changepoint_output
from skchange.costs.cost_factory import cost_factory
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_larger_than


@njit
def get_changepoints(prev_cpts: np.ndarray) -> list:
    changepoints = []
    i = len(prev_cpts) - 1
    while i >= 0:
        cpt_i = prev_cpts[i]
        changepoints.append(i)
        i = cpt_i
    return changepoints[:0:-1]  # Remove the artifical changepoint at the last index.


@njit
def run_pelt(
    X: np.ndarray, cost_func, cost_init_func, penalty, min_segment_length
) -> Tuple[np.ndarray, list]:
    params = cost_init_func(X)
    n = len(X)

    starts = np.array((), dtype=np.int64)  # Evolving set of admissible segment starts.
    init_starts = np.zeros(min_segment_length - 1, dtype=np.int64)
    init_ends = np.arange(min_segment_length - 1)
    opt_cost = np.zeros(n + 1) - penalty
    opt_cost[1:min_segment_length] = cost_func(params, init_starts, init_ends)

    # Store the previous changepoint for each t.
    # Used to get the final set of changepoints after the loop.
    prev_cpts = np.repeat(-1, n)

    ts = np.arange(min_segment_length - 1, n).reshape(-1, 1)
    for t in ts:
        starts = np.concatenate((starts, t - min_segment_length + 1))
        ends = np.repeat(t, len(starts))
        candidate_opt_costs = (
            opt_cost[starts] + cost_func(params, starts, ends) + penalty
        )
        argmin = np.argmin(candidate_opt_costs)
        opt_cost[t + 1] = candidate_opt_costs[argmin]
        prev_cpts[t] = starts[argmin] - 1

        # Trimming the admissible starts set
        starts = starts[candidate_opt_costs - penalty <= opt_cost[t]]

    return opt_cost[1:], get_changepoints(prev_cpts)


class Pelt(BaseSeriesAnnotator):
    """Pruned exact linear time changepoint detection.

    An efficient implementation of the PELT algorithm [1]_ for changepoint detection.

    Parameters
    ----------
    cost : str or callable, optional (default="mean")
        Cost function to use for changepoint detection.
        * If "mean", the Gaussian mean likelihood cost is used,
        * ...
    penalty_scale : float, optional (default=2.0)
        Scaling factor for the penalty. The penalty is set to
        'penalty_scale * 2 * p * np.log(n)', where 'n' is the sample size
        and 'p' is the number of variables. If None, the penalty is tuned on the data
        input to .fit() (not supported yet).
    min_segment_length : int, optional (default=2)
        Minimum length of a segment.
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is an
        outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a value
        belongs to.


    References
    ----------
    .. [1] Killick, R., Fearnhead, P., & Eckley, I. A. (2012). Optimal detection of
    changepoints with a linear computational cost. Journal of the American Statistical
    Association, 107(500), 1590-1598.

    Examples
    --------
    from skchange.change_detectors.pelt import Pelt
    from skchange.datasets.generate import generate_teeth_data

    df = generate_teeth_data(n_segments=2, mean=10, segment_length=10000, p=5)
    detector = Pelt()
    detector.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        cost: Union[str, Callable] = "mean",
        penalty_scale: Optional[float] = 2.0,
        min_segment_length: int = 2,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.cost = cost
        self.penalty_scale = penalty_scale
        self.min_segment_length = min_segment_length
        super().__init__(fmt=fmt, labels=labels)
        self.cost_func, self.cost_init_func = cost_factory(self.cost)

        check_larger_than(0, penalty_scale, "penalty_scale", allow_none=True)
        check_larger_than(1, min_segment_length, "min_segment_length")

    def _tune_penalty(self, X: pd.DataFrame) -> float:
        """Tune the penalty.

        Parameters
        ----------
        X : pd.DataFrame
            Training data to tune the penalty on.

        Returns
        -------
        penalty : float
            The tuned penalty.
        """
        raise ValueError(
            "tuning of the penalty is not supported yet (penalty_scale=None)."
        )

    @staticmethod
    def get_default_penalty(n: int, p: int) -> float:
        """Get the default, BIC-penalty for PELT.

        Parameters
        ----------
        n : int
            Sample size.
        p : int
            Number of variables.

        Returns
        -------
        penalty : float
            The default penalty.
        """
        return 2 * p * np.log(n)

    def _get_penalty(self, X: pd.DataFrame) -> float:
        if self.penalty_scale is None:
            return self._tune_penalty(X)
        else:
            n = X.shape[0]
            p = X.shape[1]
            return self.penalty_scale * self.get_default_penalty(n, p)

    def _fit(self, X: Union[pd.Series, pd.DataFrame], Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the penalty of the detector.
        If `penalty_scale` is None, the penalty is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `penalty_scale` is a
        number, the penalty is set to `penalty_scale` times the default penalty
        for the detector. The default penalty depends at least on the data's shape,
        but could also depend on more parameters.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the penalty to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self
        """
        X = check_data(
            X,
            min_length=2 * self.min_segment_length,
            min_length_name="2*min_segment_length",
        )
        self.penalty_ = self._get_penalty(X)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        X = check_data(
            X,
            min_length=2 * self.min_segment_length,
            min_length_name="2*min_segment_length",
        )
        opt_costs, self.changepoints = run_pelt(
            X.values,
            self.cost_func,
            self.cost_init_func,
            self.penalty_,
            self.min_segment_length,
        )
        self.scores = pd.Series(opt_costs, index=X.index, name="score")
        return format_changepoint_output(
            self.fmt, self.labels, self.changepoints, X.index, self.scores
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"cost": "mean", "min_segment_length": 5},
        ]
        return params



"""Utility functions for anomaly detection."""

from typing import List, Tuple, Union

import numpy as np
import pandas as pd


def merge_anomalies(
    collective_anomalies: Union[
        List[Tuple[int, int]], List[Tuple[int, int, np.ndarray]]
    ] = None,
    point_anomalies: Union[
        List[int],
        List[Tuple[int, int]],
        List[Tuple[int, np.ndarray]],
        List[Tuple[int, int, np.ndarray]],
    ] = None,
) -> List[Tuple[int, int, np.ndarray]]:
    """Merge collective and point anomalies into a single list of intervals.

    Parameters
    ----------
    collective_anomalies : list, optional (default=None)
        List of tuples containing inclusive start and end indices of collective
        anomalies.
    point_anomalies : list, optional (default=None)
        List of point anomaly indices.

    Returns
    -------
    list
        List of tuples containing inclusive start and end indices of collective
        anomalies and point anomalies.
    """
    if collective_anomalies is None and point_anomalies is None:
        raise ValueError(
            "Either collective_anomalies or point_anomalies must be given."
        )

    anomalies = []
    if collective_anomalies:
        anomalies += collective_anomalies
    if point_anomalies:
        # Convert point anomalies to the same format as collective anomalies
        if isinstance(point_anomalies[0], int):
            anomalies += [(i, i) for i in point_anomalies]
        elif len(point_anomalies[0]) == 2 and isinstance(
            point_anomalies[0][-1], np.ndarray
        ):
            anomalies += [(i, i, components) for (i, components) in point_anomalies]
        else:
            anomalies += point_anomalies

    anomalies = sorted(anomalies)
    return anomalies


def anomalies_to_labels(
    anomalies: List[Tuple[int, int]], n: int, p: int = None
) -> np.ndarray:
    """Convert anomaly indices to labels.

    Parameters
    ----------
    anomalies : list
        List of tuples containing inclusive start and end indices of  collective
        anomalies and point anomalies.
    n : int
        Sample size.
    p : int
        Dimensionality of the data input to the anomaly detector.

    Returns
    -------
    np.ndarray
        Array of labels, where 0 is the normal class, and 1, 2, ... are labels for each
        distinct collective and/or point_anomaly.
    """
    labels = np.zeros(n, dtype=int) if p is None else np.zeros((n, p), dtype=int)
    if len(anomalies) == 0:
        return labels

    if len(anomalies[0]) == 2:
        for i, (start, end) in enumerate(anomalies):
            labels[start : end + 1] = i + 1
    elif len(anomalies[0]) == 3:
        # Multivariate
        for i, (start, end, components) in enumerate(anomalies):
            labels[start : end + 1, components] = i + 1
    return labels


def format_anomaly_output(
    fmt: str,
    labels: str,
    X_index: pd.Index,
    collective_anomalies: List[tuple] = None,
    point_anomalies: List[tuple] = None,
    scores: Union[pd.Series, pd.DataFrame] = None,
) -> pd.Series:
    """Format the predict method output of change detectors.

    Parameters
    ----------
    fmt : str
        Format of the output. Either "sparse" or "dense".
    labels : str
        Labels of the output. Either "indicator", "score" or "int_label".
    X_index : pd.Index
        Index of the input data.
    collective_anomalies : list, optional (default=None)
        List of tuples containing inclusive start and end indices of collective
        anomalies.
    point_anomalies : list, optional (default=None)
        List of point anomaly indices.
    scores : pd.Series or pd.DataFrame, optional (default=None)
        Series or DataFrame of scores. If Series, it must be named 'score', and if
        DataFrame, it must have a column named 'score'.

    Returns
    -------
    pd.Series
        Either a sparse or dense pd.Series of boolean labels, integer labels or scores.
    """
    n = X_index.size
    anomalies = merge_anomalies(collective_anomalies, point_anomalies)
    if labels == "int_label":
        if fmt == "dense":
            anomaly_labels = anomalies_to_labels(anomalies, n)
            out = pd.Series(anomaly_labels, index=X_index, name="int_label", dtype=int)
        elif fmt == "sparse":
            out = pd.DataFrame(anomalies, columns=["start", "end"])
    elif labels == "indicator":
        if fmt == "dense":
            anomaly_labels = anomalies_to_labels(anomalies, n)
            out = pd.Series(anomaly_labels > 0, index=X_index, name="indicator")
        elif fmt == "sparse":
            out = pd.DataFrame(anomalies, columns=["start", "end"])
    elif labels == "score":
        # There is no sparse version of 'score'.
        # The scores are formatted in each class' _predict method, as what is a good
        # format for the scores is method dependent.
        out = scores
    return out


def format_multivariate_anomaly_output(
    fmt: str,
    labels: str,
    X_index: pd.Index,
    X_columns: pd.Index,
    collective_anomalies: List[dict] = None,
    point_anomalies: List[dict] = None,
    scores: Union[pd.Series, pd.DataFrame] = None,
) -> pd.Series:
    """Format the predict method output of change detectors.

    Parameters
    ----------
    fmt : str
        Format of the output. Either "sparse" or "dense".
    labels : str
        Labels of the output. Either "indicator", "score" or "int_label".
    X_index : pd.Index
        Index of the input data.
    X_columns : pd.Index
        Columns of the input data.
    collective_anomalies : list, optional (default=None)
        List of tuples containing inclusive start and end indices of collective
        anomalies.
    point_anomalies : list, optional (default=None)
        List of point anomaly indices.
    scores : pd.Series or pd.DataFrame, optional (default=None)
        Series or DataFrame of scores. If Series, it must be named 'score', and if
        DataFrame, it must have a column named 'score'.

    Returns
    -------
    pd.Series
        Either a sparse or dense pd.Series of boolean labels, integer labels or scores.
    """
    n = X_index.size
    p = X_columns.size
    anomalies = merge_anomalies(collective_anomalies, point_anomalies)
    if labels == "int_label":
        if fmt == "dense":
            anomaly_labels = anomalies_to_labels(anomalies, n, p)
            out = pd.DataFrame(
                anomaly_labels, index=X_index, columns=X_columns, dtype=int
            )
        elif fmt == "sparse":
            out = pd.DataFrame(anomalies, columns=["start", "end", "components"])
    elif labels == "indicator":
        if fmt == "dense":
            anomaly_labels = anomalies_to_labels(anomalies, n, p)
            out = pd.DataFrame(anomaly_labels > 0, index=X_index, columns=X_columns)
        elif fmt == "sparse":
            out = pd.DataFrame(anomalies, columns=["start", "end", "components"])
    elif labels == "score":
        # There is no sparse version of 'score'.
        # The scores are formatted in each class' _predict method, as what is a good
        # format for the scores is method dependent.
        out = scores
    return out



"""The Moving Score algorithm for multiple changepoint detection."""

__author__ = ["mtveten"]
__all__ = ["Moscore"]

from typing import Callable, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numba import njit
from sktime.annotation.base import BaseSeriesAnnotator

from skchange.change_detectors.utils import format_changepoint_output
from skchange.scores.score_factory import score_factory
from skchange.utils.numba.general import where
from skchange.utils.validation.data import check_data
from skchange.utils.validation.parameters import check_in_interval, check_larger_than


@njit
def get_moscore_changepoints(
    moscores: np.ndarray, threshold: float, min_detection_interval: int
) -> list:
    detection_intervals = where(moscores > threshold)
    changepoints = []
    for interval in detection_intervals:
        start = interval[0]
        end = interval[1]
        if end - start + 1 >= min_detection_interval:
            cpt = np.argmax(moscores[start : end + 1]) + start
            changepoints.append(cpt)
    return changepoints


@njit
def moscore_transform(
    X: np.ndarray,
    score_f: Callable,
    score_init_f: Callable,
    bandwidth: int,
) -> Tuple[list, np.ndarray]:
    n = len(X)
    splits = np.arange(bandwidth - 1, n - bandwidth)
    starts = splits - bandwidth + 1
    ends = splits + bandwidth
    params = score_init_f(X)
    scores = np.zeros(n)
    scores[splits] = score_f(params, starts, ends, splits)
    return scores


class Moscore(BaseSeriesAnnotator):
    """Moving score algorithm for multiple changepoint detection.

    A generalized version of the MOSUM (moving sum) algorithm [1]_ for changepoint
    detection. It runs a test statistic for a single changepoint at the midpoint in a
    moving window of length `2*bandwidth` over the data. Efficently implemented using
    numba.

    Parameters
    ----------
    score: str, Tuple[Callable, Callable], optional (default="mean")
        Test statistic to use for changepoint detection.
        * If "mean", the difference-in-mean statistic is used,
        * If "var", the difference-in-variance statistic is used,
        * If a tuple, it must contain two functions: The first function is the scoring
        function, which takes in the output of the second function as its first
        argument, and start, end and split indices as the second, third and fourth
        arguments. The second function is the initializer, which precomputes quantities
        that should be precomputed. See skchange/scores/score_factory.py for examples.
    bandwidth : int, optional (default=30)
        The bandwidth is the number of samples on either side of a candidate
        changepoint. The minimum bandwidth depends on the
        test statistic. For "mean", the minimum bandwidth is 1.
    threshold_scale : float, optional (default=2.0)
        Scaling factor for the threshold. The threshold is set to
        'threshold_scale * default_threshold', where the default threshold depends on
        the number of samples, the number of variables, `bandwidth` and `level`.
        If None, the threshold is tuned on the data input to .fit().
    level : float, optional (default=0.01)
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the changepoint score on the training data. For this to be correct, the
        training data must contain no changepoints. If `threshold_scale` is a number,
        `level` is used in the default threshold, _before_ scaling.
    min_detection_interval : int, optional (default=1)
        Minimum number of consecutive scores above the threshold to be considered a
        changepoint. Must be between 1 and `bandwidth`/2.
    fmt : str {"dense", "sparse"}, optional (default="sparse")
        Annotation output format:
        * If "sparse", a sub-series of labels for only the outliers in X is returned,
        * If "dense", a series of labels for all values in X is returned.
    labels : str {"indicator", "score", "int_label"}, optional (default="int_label")
        Annotation output labels:
        * If "indicator", returned values are boolean, indicating whether a value is an
        outlier,
        * If "score", returned values are floats, giving the outlier score.
        * If "int_label", returned values are integer, indicating which segment a value
        belongs to.

    References
    ----------
    .. [1] Eichinger, B., & Kirch, C. (2018). A MOSUM procedure for the estimation of
    multiple random change points.

    Examples
    --------
    from skchange.change_detectors.moscore import Moscore
    from skchange.datasets.generate import generate_teeth_data

    df = generate_teeth_data(n_segments=2, mean=10, segment_length=100000, p=5)
    detector = Moscore()
    detector.fit_predict(df)
    """

    _tags = {
        "capability:missing_values": False,
        "capability:multivariate": True,
        "fit_is_empty": False,
    }

    def __init__(
        self,
        score: Union[str, Tuple[Callable, Callable]] = "mean",
        bandwidth: int = 30,
        threshold_scale: Optional[float] = 2.0,
        level: float = 0.01,
        min_detection_interval: int = 1,
        fmt: str = "sparse",
        labels: str = "int_label",
    ):
        self.score = score
        self.bandwidth = bandwidth
        self.threshold_scale = threshold_scale  # Just holds the input value.
        self.level = level
        self.min_detection_interval = min_detection_interval
        super().__init__(fmt=fmt, labels=labels)
        self.score_f, self.score_init_f = score_factory(self.score)

        check_larger_than(1, self.bandwidth, "bandwidth")
        check_larger_than(0, threshold_scale, "threshold_scale", allow_none=True)
        check_larger_than(0, self.level, "level")
        check_in_interval(
            pd.Interval(1, self.bandwidth / 2 - 1, closed="both"),
            self.min_detection_interval,
            "min_detection_interval",
        )

    def _tune_threshold(self, X: pd.DataFrame) -> float:
        """Tune the threshold for the Moscore algorithm.

        The threshold is set to the (1-`level`)-quantile of the score on the training
        data `X`. For this to be correct, the training data must contain no
        changepoints.

        TODO: Find the threshold given an input number `k` of "permitted" changepoints
        in the training data. This can be achieved by filtering out the top `k` peaks
        of the score.

        Parameters
        ----------
        X : pd.DataFrame
            Training data to tune the threshold on.
        """
        scores = moscore_transform(
            X.values,
            self.score_f,
            self.score_init_f,
            self.bandwidth,
        )
        tuned_threshold = np.quantile(scores, 1 - self.level)
        return tuned_threshold

    @staticmethod
    def get_default_threshold(
        n: int, p: int, bandwidth: int, level: float = 0.01
    ) -> float:
        """Get the default threshold for the Moscore algorithm.

        It is the asymptotic critical value of the univariate 'mean' test statitic,
        multiplied by `p` to account for the multivariate case.

        Parameters
        ----------
        n : int
            Sample size.
        p : int
            Number of variables.
        bandwidth : int
            Bandwidth of the Moscore algorithm.
        level : float, optional (default=0.01)
            Significance level for the test statistic.

        Returns
        -------
        threshold : float
            Threshold for the Moscore algorithm.
        """
        u = n / bandwidth
        a = np.sqrt(2 * np.log(u))
        b = (
            2 * np.log(u)
            + 1 / 2 * np.log(np.log(u))
            + np.log(3 / 2)
            - 1 / 2 * np.log(np.pi)
        )
        c = -np.log(np.log(1 / np.sqrt(1 - level)))
        # TODO: Check if it's correct to multiply by p.
        return p * (b + c) / a

    def _get_threshold(self, X: pd.DataFrame) -> float:
        if self.threshold_scale is None:
            return self._tune_threshold(X)
        else:
            n = X.shape[0]
            p = X.shape[1]
            return self.threshold_scale * self.get_default_threshold(
                n, p, self.bandwidth, self.level
            )

    def _fit(self, X: pd.DataFrame, Y: Optional[pd.DataFrame] = None):
        """Fit to training data.

        Sets the threshold of the detector.
        If `threshold_scale` is None, the threshold is set to the (1-`level`)-quantile
        of the change/anomaly scores on the training data. For this to be correct,
        the training data must contain no changepoints. If `threshold_scale` is a
        number, the threshold is set to `threshold_scale` times the default threshold
        for the detector. The default threshold depends at least on the data's shape,
        but could also depend on more parameters.
        In the case of the Moscore algorithm, the default threshold depends on the
        sample size, the number of variables, `bandwidth` and `level`.

        Parameters
        ----------
        X : pd.DataFrame
            training data to fit the threshold to.
        Y : pd.Series, optional
            Does nothing. Only here to make the fit method compatible with sktime
            and scikit-learn.

        Returns
        -------
        self : returns a reference to self

        State change
        ------------
        Creates the threshold_ attribute.
        """
        X = check_data(
            X,
            min_length=2 * self.bandwidth,
            min_length_name="2*bandwidth",
        )
        self.threshold_ = self._get_threshold(X)
        return self

    def _predict(self, X: Union[pd.DataFrame, pd.Series]) -> pd.Series:
        """Create annotations on test/deployment data.

        Parameters
        ----------
        X : pd.DataFrame - data to annotate, time series

        Returns
        -------
        Y : pd.Series - annotations for sequence X
            exact format depends on annotation type
        """
        X = check_data(
            X,
            min_length=2 * self.bandwidth,
            min_length_name="2*bandwidth",
        )
        scores = moscore_transform(
            X.values,
            self.score_f,
            self.score_init_f,
            self.bandwidth,
        )
        self.changepoints = get_moscore_changepoints(
            scores, self.threshold_, self.min_detection_interval
        )
        self.scores = pd.Series(scores, index=X.index, name="score")
        return format_changepoint_output(
            self.fmt, self.labels, self.changepoints, X.index, self.scores
        )

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            There are currently no reserved values for annotators.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = [
            {"score": "mean", "bandwidth": 10},
        ]
        return params



"""Gaussian mean likelihood cost function for change point detection."""

from typing import Tuple

import numpy as np
from numba import njit

from skchange.utils.numba.general import col_repeat
from skchange.utils.numba.stats import col_cumsum


@njit(cache=True)
def init_mean_cost(X: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Precompute sums and weights for mean_cost.

    Parameters
    ----------
    X : np.ndarray
        2D array.

    Returns
    -------
    sums : np.ndarray
        Cumulative sums of X.
    sums2 : np.ndarray
        Cumulative sums of X**2.
    weights : np.ndarray
        Weights for sums2 in the cost calculation.
    """
    n = X.shape[0]
    p = X.shape[1]

    # 0.0 as first row to make calculations work also for start = 0
    sums = np.zeros((n + 1, p))
    sums[1:] = col_cumsum(X)
    sums2 = np.zeros((n + 1, p))
    sums2[1:] = col_cumsum(X**2)
    weights = col_repeat(np.arange(0, n + 1), p)
    return sums, sums2, weights


@njit(cache=True)
def mean_cost(
    precomputed_params: Tuple[np.ndarray, np.ndarray, np.ndarray],
    starts: np.ndarray,
    ends: np.ndarray,
) -> np.ndarray:
    """Calculate the Gaussian mean likelihood cost for each segment.

    Parameters
    ----------
    precomputed_params : Tuple[np.ndarray, np.ndarray, np.ndarray]
        Precomputed parameters from init_mean_cost.
    starts : np.ndarray
        Start indices of the segments.
    ends : np.ndarray
        End indices of the segments.

    Returns
    -------
    costs : np.ndarray
        Costs for each segment.
    """
    sums, sums2, weights = precomputed_params
    partial_sums = sums[ends + 1] - sums[starts]
    partial_sums2 = sums2[ends + 1] - sums2[starts]
    weights = weights[ends - starts + 1]
    costs = np.sum(partial_sums2 - partial_sums**2 / weights, axis=1)
    return costs



import numpy as np

from skchange.change_detectors.pelt import Pelt
from skchange.costs.mean_cost import init_mean_cost, mean_cost
from skchange.datasets.generate import generate_teeth_data
from skchange.utils.benchmarking.profiler import Profiler

# Simple univariate example
df = generate_teeth_data(n_segments=2, mean=10, segment_length=100, p=1, random_state=2)
detector = Pelt(min_segment_length=1)
detector.fit_predict(df)


# Profiling
n = int(1e6)
df = generate_teeth_data(n_segments=1, mean=0, segment_length=n, p=1)
detector = Pelt()
profiler = Profiler()
profiler.start()
detector.fit_predict(df)
profiler.stop()


# Various unit tests
df = generate_teeth_data(n_segments=3, mean=10, segment_length=5, p=2)
precomputed_params = init_mean_cost(df.values)
mean_cost(precomputed_params, starts=np.array([1, 2, 3]), ends=np.array([3, 4, 5]))
detector = Pelt()
detector.fit_predict(df)
